"""
Perform prompt-completion classification using a model which can be loaded via
``llama_cpp.Llama``.

You probably just want the :func:`predict` function :-)

The examples below use a 6 MB model to quickly demonstrate functionality. To download
it, first install ``huggingface-hub`` if you don't have it already::

    pip install huggingface-hub

And then download `the model
<https://huggingface.co/aladar/TinyLLama-v0-GGUF/blob/main/TinyLLama-v0.Q8_0.gguf>`_ (to
your current working directory)::

    huggingface-cli download \\
    aladar/TinyLLama-v0-GGUF \\
    TinyLLama-v0.Q8_0.gguf \\
    --local-dir . \\
    --local-dir-use-symlinks False
"""

from __future__ import annotations
from contextlib import contextmanager
from typing import cast, Literal, Sequence

from llama_cpp import Llama
import numpy as np
import numpy.typing as npt

from cappr.utils import classify
from cappr.utils._batch import ProgressBar
from cappr import Example
from cappr.llama_cpp import _utils


@classify._token_logprobs
def token_logprobs(
    texts: str | Sequence[str],
    model: Llama,
    end_of_prompt: Literal[" ", ""] = " ",
    show_progress_bar: bool | None = None,
    add_bos: bool = False,
    **kwargs,
) -> list[float] | list[list[float]]:
    """
    For each text, compute each token's log-probability conditional on all previous
    tokens in the text.

    Parameters
    ----------
    texts : str | Sequence[str]
        input text(s)
    model : Llama
        a Llama CPP model
    end_of_prompt : Literal[' ', ''], optional
        This string gets added to the beginning of each text. It's important to set this
        if you're using the discount feature. Otherwise, set it to "". By default " "
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 texts
    add_bos : bool, optional
        whether or not to add a beginning-of-sentence token to each text in `texts` if
        the tokenizer has a beginning-of-sentence token, by default False

    Returns
    -------
    log_probs : list[float] | list[list[float]]

        If `texts` is a string, then a 1-D list is returned: `log_probs[token_idx]` is
        the log-probability of the token at `token_idx` of `texts` conditional on all
        previous tokens in `texts`.

        If `texts` is a sequence of strings, then a 2-D list is returned:
        `log_probs[text_idx][token_idx]` is the log-probability of the token at
        `token_idx` of `texts[text_idx]` conditional on all previous tokens in
        `texts[text_idx]`.

    Warning
    -------
    Set `end_of_prompt="", add_bos=True` unless you're using the discount feature.

    Note
    ----
    For each text, the first token's log-probability is always ``None`` because no
    autoregressive LM directly estimates the marginal probability of a token.

    Raises
    ------
    TypeError
        if `texts` is not a sequence
    ValueError
        if `texts` is empty
    """
    if not _utils.does_tokenizer_need_prepended_space(model):
        end_of_prompt = ""
    texts = [end_of_prompt + text for text in texts]
    # Loop through completions, b/c llama cpp currently doesn't support batch inference
    # Note: we could instead run logits_to_log_probs over a batch to save a bit of time,
    # but that'd cost more memory
    log_probs = []
    first_token_log_prob = [None]
    with _utils.set_up_model(model):
        for text in ProgressBar(
            texts, show_progress_bar=show_progress_bar, desc="marginal log-probs"
        ):
            input_ids = model.tokenize(text.encode(), add_bos=add_bos)
            model.reset()  # clear the model's KV cache and logits
            model.eval(input_ids)
            log_probs_text: list[float] = _utils.logits_to_log_probs(
                np.array(model.eval_logits),
                np.array(input_ids),
                input_ids_start_idx=1,  # this log-prob is in the prev logit
                logits_end_idx=-1,
            ).tolist()
            log_probs.append(first_token_log_prob + log_probs_text)
        model.reset()
    return log_probs


########################################################################################
###################################### KV caching ######################################
########################################################################################


def cache_model(model: Llama, prefix: str) -> Llama:
    """
    Caches the model so that every future computation with it starts with `prefix`. As a
    result, computations with this model are faster.

    Use this function instead of the context manager :func:`cache` to keep the cache for
    future computations, including those outside of a context.

    Parameters
    ----------
    model : Llama
        a Llama CPP model
    prefix : str
        prefix for all future strings that will be processed, e.g., a string containing
        shared prompt instructions, or a string containing instructions and exemplars
        for few-shot prompting

    Note
    ----
    When inputting this model to a function from this module, set the function's
    `reset_model=False`.

    Example
    -------
    Usage with :func:`predict_proba`::

        import numpy as np
        from llama_cpp import Llama
        from cappr.llama_cpp.classify import cache_model, predict_proba

        # Load model
        # The top of this page has instructions to download this model
        model_path = "./TinyLLama-v0.Q8_0.gguf"
        model = Llama(model_path, verbose=False)

        # Create data
        prompt_prefix = "Once upon a time,"
        prompts = ["there was", "in a land far far, far away,"]
        completions = [
            "a llama in pajamas",
            "an alpaca in Havana",
        ]

        # Compute
        model = cache_model(model, prompt_prefix)
        # Always set reset_model=False
        pred_probs = predict_proba(
            prompts, completions, model, reset_model=False
        )

        # The above computation is equivalent to this one:
        prompts_full = [prompt_prefix + " " + prompt for prompt in prompts]
        pred_probs_wo_cache = predict_proba(
            prompts_full, completions, model, reset_model=True
        )
        assert np.allclose(pred_probs, pred_probs_wo_cache)
    """
    if prefix:
        # W/o the if condition, we'd eval on <bos> if there's no prefix and no cache
        input_ids_prefix = model.tokenize(prefix.encode(), add_bos=model.n_tokens == 0)
        with _utils.set_up_model(model):
            model.eval(input_ids_prefix)
    return model


@contextmanager
def cache(model: Llama, prefix: str, reset_model: bool = True):
    """
    In this context, every prompt processed by the `model` starts with `prefix + " "`.
    As a result, computations in this context are faster.

    Parameters
    ----------
    model : Llama
        a Llama CPP model
    prefix : str
        prefix for all strings that will be processed in this context, e.g., a string
        containing shared prompt instructions, or a string containing instructions and
        exemplars for few-shot prompting
    reset_model : bool, optional
        whether or not to reset the model's KV cache and logits. Set this to False when
        you're in a :func:`cache` context. By default, True

    Note
    ----
    In this context, when using a function from this module, set their
    `reset_model=False`.

    Example
    -------
    Usage with :func:`predict_proba`::

        import numpy as np
        from llama_cpp import Llama
        from cappr.llama_cpp.classify import cache, predict_proba

        # Load model
        # The top of this page has instructions to download this model
        model_path = "./TinyLLama-v0.Q8_0.gguf"
        model = Llama(model_path, verbose=False)

        # Create data
        prompt_prefix = "Once upon a time,"
        prompts = ["there was", "in a land far far, far away,"]
        completions = [
            "a llama in pajamas",
            "an alpaca in Havana",
        ]

        # Compute
        with cache(model, prompt_prefix):
            # Always set reset_model=False
            pred_probs = predict_proba(
                prompts, completions, model, reset_model=False
            )

        # The above computation is equivalent to this one:
        prompts_full = [prompt_prefix + " " + prompt for prompt in prompts]
        pred_probs_wo_cache = predict_proba(
            prompts_full, completions, model, reset_model=True
        )
        assert np.allclose(pred_probs, pred_probs_wo_cache)
    """
    if reset_model:
        model.reset()
    n_tokens = model.n_tokens
    try:
        yield cache_model(model, prefix)
    finally:
        model.n_tokens = n_tokens


########################################################################################
############################## Logprobs from cached model ##############################
########################################################################################


def _log_probs_conditional_prompt(
    prompt: str,
    completions: Sequence[str],
    model: Llama,
    end_of_prompt: Literal[" ", ""],
) -> list[list[float]]:
    # Prepend whitespaces if the tokenizer or context call for it
    # TODO: put this in the context manager? Little weird
    if not _utils.does_tokenizer_need_prepended_space(model):
        start_of_prompt = ""
        end_of_prompt = ""
    else:
        in_cache_context = model.n_tokens > 0
        start_of_prompt = " " if in_cache_context else ""
    prompt = start_of_prompt + prompt
    completions = [end_of_prompt + completion for completion in completions]

    # Cache the prompt's KVs and logits
    with cache(model, prompt, reset_model=False):
        num_tokens_prompt = model.n_tokens
        # Tokenize completions to determine whether or not we can do the single-token
        # optimization
        #
        # For Llama (and probably others) we don't want the completions to start w/ a
        # bos token <s> b/c we need to mimic sending the prompt + completion together
        # For example, if 'a b' is the prompt and 'c' is the completion, the encoding
        # should correspond to '<s> a b c' not '<s> a b <s> c'
        input_ids_completions = [
            model.tokenize(completion.encode(), add_bos=False)
            for completion in completions
        ]
        if all(
            len(input_ids_completion) == 1
            for input_ids_completion in input_ids_completions
        ):
            # Single-token optimization
            prompt_next_token_log_probs = _utils.log_softmax(
                np.array(model.eval_logits[-1])
            )
            return [
                [prompt_next_token_log_probs[input_ids_completion[0]]]
                for input_ids_completion in input_ids_completions
            ]
        # Loop through completions, b/c llama cpp currently doesn't support batch
        # inference
        log_probs_completions: list[list[float]] = []
        with _utils.set_up_model(model):
            for input_ids_completion in input_ids_completions:
                # Given the prompt, compute next-token logits for each completion token
                model.eval(input_ids_completion)
                # Logits -> log-probs. We need the prompt's last token's logits b/c it
                # contains the first completion token's log-prob. But we don't need the
                # last completion token's next-token logits ofc. Also, it's
                # num_tokens_prompt - 1 b/c of 0-indexing
                logits_completion = np.array(model.eval_logits)[
                    num_tokens_prompt - 1 : -1
                ]
                log_probs_completion: list[float] = _utils.logits_to_log_probs(
                    logits_completion, np.array(input_ids_completion)
                ).tolist()
                log_probs_completions.append(log_probs_completion)
                # Reset the model's KV cache to the prompt. Without this line, the cache
                # would include this completion's KVs, which is mega wrong
                model.n_tokens = num_tokens_prompt
        return log_probs_completions


########################################################################################
#################################### Implementation ####################################
########################################################################################


@classify._log_probs_conditional
def log_probs_conditional(
    prompts: str | Sequence[str],
    completions: Sequence[str],
    model: Llama,
    end_of_prompt: Literal[" ", ""] = " ",
    show_progress_bar: bool | None = None,
    reset_model: bool = True,
    **kwargs,
) -> list[list[float]] | list[list[list[float]]]:
    """
    Log-probabilities of each completion token conditional on each prompt and previous
    completion tokens.

    Parameters
    ----------
    prompts : str | Sequence[str]
        string(s), where, e.g., each contains the text you want to classify
    completions : Sequence[str]
        strings, where, e.g., each one is the name of a class which could come after a
        prompt
    model : Llama
        a Llama CPP model
    end_of_prompt : Literal[' ', ''], optional
        whitespace or empty string to join prompt and completion, by default whitespace
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 prompts
    reset_model : bool, optional
        whether or not to reset the model's KV cache and logits. Set this to False when
        you're in a :func:`cache` context. By default, True

    Returns
    -------
    log_probs_completions : list[list[float]] | list[list[list[float]]]

        If `prompts` is a string, then a 2-D list is returned:
        `log_probs_completions[completion_idx][completion_token_idx]` is the
        log-probability of the completion token in `completions[completion_idx]`,
        conditional on `prompt` and previous completion tokens.

        If `prompts` is a sequence of strings, then a 3-D list is returned:
        `log_probs_completions[prompt_idx][completion_idx][completion_token_idx]` is the
        log-probability of the completion token in `completions[completion_idx]`,
        conditional on `prompts[prompt_idx]` and previous completion tokens.

    Note
    ----
    To efficiently aggregate `log_probs_completions`, use
    :func:`cappr.utils.classify.agg_log_probs`.

    Example
    -------
    Here we'll use single characters (which are single tokens) to more clearly
    demonstrate what this function does::

        from llama_cpp import Llama
        from cappr.llama_cpp.classify import log_probs_conditional

        # Load model
        # The top of this page has instructions to download this model
        model_path = "./TinyLLama-v0.Q8_0.gguf"
        model = Llama(model_path, verbose=False)

        # Create data
        prompts = ["x y", "a b c"]
        completions = ["z", "d e"]

        # Compute
        log_probs_completions = log_probs_conditional(
            prompts, completions, model
        )

        # Outputs (rounded) next to their symbolic representation

        print(log_probs_completions[0])
        # [[-12.8],        [[log Pr(z | x, y)],
        #  [-10.8, -10.7]]  [log Pr(d | x, y),    log Pr(e | x, y, d)]]

        print(log_probs_completions[1])
        # [[-9.5],        [[log Pr(z | a, b, c)],
        #  [-9.9, -10.0]]  [log Pr(d | a, b, c), log Pr(e | a, b, c, d)]]
    """
    if reset_model:
        model.reset()
    log_probs_completions = [
        _log_probs_conditional_prompt(prompt, completions, model, end_of_prompt)
        for prompt in ProgressBar(
            prompts, show_progress_bar=show_progress_bar, desc="conditional log-probs"
        )
    ]
    if reset_model:
        model.reset()
    return log_probs_completions


@classify._log_probs_conditional_examples
def log_probs_conditional_examples(
    examples: Example | Sequence[Example],
    model: Llama,
    show_progress_bar: bool | None = None,
    reset_model: bool = True,
) -> list[list[float]] | list[list[list[float]]]:
    """
    Log-probabilities of each completion token conditional on each prompt and previous
    completion tokens.

    Parameters
    ----------
    examples : Example | Sequence[Example]
        `Example` object(s), where each contains a prompt and its set of possible
        completions
    model : Llama
        a Llama CPP model
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 `examples`
    reset_model : bool, optional
        whether or not to reset the model's KV cache and logits. Set this to False when
        you're in a :func:`cache` context. By default, True

    Returns
    -------
    log_probs_completions : list[list[float]] | list[list[list[float]]]

        If `examples` is a :class:`cappr.Example`, then a 2-D list is returned:
        `log_probs_completions[completion_idx][completion_token_idx]` is the
        log-probability of the completion token in
        `example.completions[completion_idx]`, conditional on `example.prompt` and
        previous completion tokens.

        If `examples` is a sequence of :class:`cappr.Example` objects, then a 3-D list
        is returned:
        `log_probs_completions[example_idx][completion_idx][completion_token_idx]` is
        the log-probability of the completion token in
        `examples[example_idx].completions[completion_idx]`, conditional on
        `examples[example_idx].prompt` and previous completion tokens.

    Note
    ----
    To aggregate `log_probs_completions`, use
    :func:`cappr.utils.classify.agg_log_probs`.

    Note
    ----
    The attribute :attr:`cappr.Example.prior` is unused.

    Example
    -------
    Here we'll use single characters (which are single tokens) to more clearly
    demonstrate what this function does::

        from llama_cpp import Llama
        from cappr import Example
        from cappr.llama_cpp.classify import log_probs_conditional_examples

        # Load model
        # The top of this page has instructions to download this model
        model_path = "./TinyLLama-v0.Q8_0.gguf"
        model = Llama(model_path, verbose=False)

        # Create examples
        examples = [
            Example(prompt="x y", completions=("z", "d e")),
            Example(prompt="a b c", completions=("d e",), normalize=False),
        ]

        # Compute
        log_probs_completions = log_probs_conditional_examples(
            examples, model
        )

        # Outputs (rounded) next to their symbolic representation

        print(log_probs_completions[0])  # corresponds to examples[0]
        # [[-12.8],        [[log Pr(z | x, y)],
        #  [-10.8, -10.7]]  [log Pr(d | x, y),    log Pr(e | x, y, d)]]

        print(log_probs_completions[1])  # corresponds to examples[1]
        # [[-9.90, -10.0]] [[log Pr(d | a, b, c)], log Pr(e | a, b, c, d)]]
    """
    # examples is always a Sequence[Example] b/c of the decorator
    examples = cast(Sequence[Example], examples)
    if reset_model:
        model.reset()
    log_probs_completions = [
        _log_probs_conditional_prompt(
            example.prompt, example.completions, model, example.end_of_prompt
        )
        for example in ProgressBar(
            examples, show_progress_bar=show_progress_bar, desc="conditional log-probs"
        )
    ]
    if reset_model:
        model.reset()
    return log_probs_completions


@classify._predict_proba
def predict_proba(
    prompts: str | Sequence[str],
    completions: Sequence[str],
    model: Llama,
    end_of_prompt: Literal[" ", ""] = " ",
    prior: Sequence[float] | None = None,
    normalize: bool = True,
    discount_completions: float = 0.0,
    log_marg_probs_completions: Sequence[Sequence[float]] | None = None,
    show_progress_bar: bool | None = None,
    reset_model: bool = True,
) -> npt.NDArray[np.floating]:
    """
    Predict probabilities of each completion coming after each prompt.

    Parameters
    ----------
    prompts : str | Sequence[str]
        string(s), where, e.g., each contains the text you want to classify
    completions : Sequence[str]
        strings, where, e.g., each one is the name of a class which could come after a
        prompt
    model : Llama
        a Llama CPP model
    end_of_prompt : Literal[' ', ''], optional
        whitespace or empty string to join prompt and completion, by default whitespace
    prior : Sequence[float] | None, optional
        a probability distribution over `completions`, representing a belief about their
        likelihoods regardless of the prompt. By default, each completion in
        `completions` is assumed to be equally likely
    normalize : bool, optional
        whether or not to normalize completion-after-prompt probabilities into a
        probability distribution over completions. Set this to `False` if you'd like the
        raw completion-after-prompt probability, or you're solving a multi-label
        prediction problem. By default, True
    discount_completions : float, optional
        experimental feature: set it (e.g., 1.0 may work well) if a completion is
        consistently getting too high predicted probabilities. You could instead fudge
        the `prior`, but this hyperparameter may be easier to tune than the `prior`. By
        default 0.0
    log_marg_probs_completions : Sequence[Sequence[float]] | None, optional
        experimental feature: pre-computed log probabilities of completion tokens
        conditional on previous completion tokens (not prompt tokens). Only used if `not
        discount_completions`. Pre-compute them by passing `completions` and `model` to
        :func:`token_logprobs`. By default, if `not discount_completions`, they are
        (re-)computed
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 prompts
    reset_model : bool, optional
        whether or not to reset the model's KV cache and logits. Set this to False when
        you're in a :func:`cache` context. By default, True

    Returns
    -------
    pred_probs : npt.NDArray[np.floating]

        If `prompts` is a string, then an array with shape `len(completions),` is
        returned: `pred_probs[completion_idx]` is the model's estimate of the
        probability that `completions[completion_idx]` comes after `prompt`.

        If `prompts` is a sequence of strings, then an array with shape `(len(prompts),
        len(completions))` is returned: `pred_probs[prompt_idx, completion_idx]` is the
        model's estimate of the probability that `completions[completion_idx]` comes
        after `prompts[prompt_idx]`.

    Note
    ----
    In this function, the set of possible completions which could follow each prompt is
    the same for every prompt. If instead, each prompt could be followed by a
    *different* set of completions, then construct a sequence of :class:`cappr.Example`
    objects and pass them to :func:`predict_proba_examples`.

    Example
    -------
    Let's have our little Llama predict some story beginnings::

        from llama_cpp import Llama
        from cappr.llama_cpp.classify import predict_proba

        # Load model
        # The top of this page has instructions to download this model
        model_path = "./TinyLLama-v0.Q8_0.gguf"
        model = Llama(model_path, verbose=False)

        # Define a classification task
        prompts = ["In a hole in", "Once upon"]
        completions = ("a time", "the ground")

        # Compute
        pred_probs = predict_proba(prompts, completions, model)

        pred_probs_rounded = pred_probs.round(2)  # just for cleaner output

        # predicted probability that the ending for the clause
        # "In a hole in" is "the ground"
        print(pred_probs_rounded[0, 1]) # 0.98

        # predicted probability that the ending for the clause
        # "Once upon" is "a time"
        print(pred_probs_rounded[1, 0]) # 1.0
    """
    return log_probs_conditional(**locals())


@classify._predict_proba_examples
def predict_proba_examples(
    examples: Example | Sequence[Example],
    model: Llama,
    show_progress_bar: bool | None = None,
    reset_model: bool = True,
) -> npt.NDArray[np.floating] | list[npt.NDArray[np.floating]]:
    """
    Predict probabilities of each completion coming after each prompt.

    Parameters
    ----------
    examples : Example | Sequence[Example]
        `Example` object(s), where each contains a prompt and its set of possible
        completions
    model : Llama
        a Llama CPP model
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 `examples`
    reset_model : bool, optional
        whether or not to reset the model's KV cache and logits. Set this to False when
        you're in a :func:`cache` context. By default, True

    Returns
    -------
    pred_probs : npt.NDArray[np.floating] | list[npt.NDArray[np.floating]]

        If `examples` is an :class:`cappr.Example`, then an array with shape
        `(len(example.completions),)` is returned: `pred_probs[completion_idx]` is the
        model's estimate of the probability that `example.completions[completion_idx]`
        comes after `example.prompt`.

        If `examples` is a sequence of :class:`cappr.Example` objects, then a list with
        length `len(examples)` is returned: `pred_probs[example_idx][completion_idx]` is
        the model's estimate of the probability that
        `examples[example_idx].completions[completion_idx]` comes after
        `examples[example_idx].prompt`. If the number of completions per example is a
        constant `k`, then an array with shape `(len(examples), k)` is returned instead
        of a list of 1-D arrays.

    Example
    -------
    Some story analysis::

        from llama_cpp import Llama
        from cappr import Example
        from cappr.llama_cpp.classify import predict_proba_examples

        # Load model
        # The top of this page has instructions to download this model
        model_path = "./TinyLLama-v0.Q8_0.gguf"
        model = Llama(model_path, verbose=False)

        # Create examples
        examples = [
            Example(
                prompt="Story: I enjoyed pizza with my buddies.\\nMoral:",
                completions=("make friends", "food is yummy", "absolutely nothing"),
                prior=(2 / 5, 2 / 5, 1 / 5),
            ),
            Example(
                prompt="The child rescued the animal. The child is a",
                completions=("hero", "villain"),
            ),
        ]

        # Compute
        pred_probs = predict_proba_examples(examples, model)

        # predicted probability that the moral of the 1st story is that food is yummy
        print(pred_probs[0][1].round(2))
        # 0.72

        # predicted probability that the hero of the 2nd story is the child
        print(pred_probs[1][0].round(2))
        # 0.95
    """
    return log_probs_conditional_examples(**locals())


@classify._predict
def predict(
    prompts: str | Sequence[str],
    completions: Sequence[str],
    model: Llama,
    end_of_prompt: Literal[" ", ""] = " ",
    prior: Sequence[float] | None = None,
    discount_completions: float = 0.0,
    log_marg_probs_completions: Sequence[Sequence[float]] | None = None,
    show_progress_bar: bool | None = None,
    reset_model: bool = True,
) -> str | list[str]:
    """
    Predict which completion is most likely to follow each prompt.

    Parameters
    ----------
    prompts : str | Sequence[str]
        string(s), where, e.g., each contains the text you want to classify
    completions : Sequence[str]
        strings, where, e.g., each one is the name of a class which could come after a
        prompt
    model : Llama
        a Llama CPP model
    end_of_prompt : Literal[' ', ''], optional
        whitespace or empty string to join prompt and completion, by default whitespace
    prior : Sequence[float] | None, optional
        a probability distribution over `completions`, representing a belief about their
        likelihoods regardless of the prompt. By default, each completion in
        `completions` is assumed to be equally likely
    discount_completions : float, optional
        experimental feature: set it to >0.0 (e.g., 1.0 may work well) if a completion
        is consistently getting over-predicted. You could instead fudge the `prior`, but
        this hyperparameter may be easier to tune than the `prior`. By default 0.0
    log_marg_probs_completions : Sequence[Sequence[float]] | None, optional
        experimental feature: pre-computed log probabilities of completion tokens
        conditional on previous completion tokens (not prompt tokens). Only used if `not
        discount_completions`. Pre-compute them by passing `completions` and `model` to
        :func:`token_logprobs`. By default, if `not discount_completions`, they are
        (re-)computed
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 prompts
    reset_model : bool, optional
        whether or not to reset the model's KV cache and logits. Set this to False when
        you're in a :func:`cache` context. By default, True

    Returns
    -------
    preds : str | list[str]

        If `prompts` is a string, then the completion from `completions` which is
        predicted to most likely follow `prompt` is returned.

        If `prompts` is a sequence of strings, then a list with length `len(prompts)` is
        returned. `preds[prompt_idx]` is the completion in `completions` which is
        predicted to follow `prompts[prompt_idx]`.

    Note
    ----
    In this function, the set of possible completions which could follow each prompt is
    the same for every prompt. If instead, each prompt could be followed by a
    *different* set of completions, then construct a sequence of :class:`cappr.Example`
    objects and pass them to :func:`predict_examples`.

    Example
    -------
    Let's have our little Llama predict some story beginnings::

        from llama_cpp import Llama
        from cappr.llama_cpp.classify import predict

        # Load model
        # The top of this page has instructions to download this model
        model_path = "./TinyLLama-v0.Q8_0.gguf"
        model = Llama(model_path, verbose=False)

        # Define a classification task
        prompts = ["In a hole in", "Once upon"]
        completions = ("a time", "the ground")

        # Compute
        preds = predict(prompts, completions, model)

        # Predicted ending for the first clause: "In a hole in"
        print(preds[0])
        # the ground

        # Predicted ending for the first clause: "Once upon"
        print(preds[1])
        # a time
    """
    return predict_proba(**locals())


@classify._predict_examples
def predict_examples(
    examples: Example | Sequence[Example],
    model: Llama,
    show_progress_bar: bool | None = None,
    reset_model: bool = True,
) -> str | list[str]:
    """
    Predict which completion is most likely to follow each prompt.

    Parameters
    ----------
    examples : Example | Sequence[Example]
        `Example` object(s), where each contains a prompt and its set of possible
        completions
    model : Llama
        a Llama CPP model
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 `examples`
    reset_model : bool, optional
        whether or not to reset the model's KV cache and logits. Set this to False when
        you're in a :func:`cache` context. By default, True

    Returns
    -------
    preds : str | list[str]

        If `examples` is an :class:`cappr.Example`, then the completion from
        `example.completions` which is predicted to most likely follow `example.prompt`
        is returned.

        If `examples` is a sequence of :class:`cappr.Example` objects, then a list with
        length `len(examples)` is returned: `preds[example_idx]` is the completion in
        `examples[example_idx].completions` which is predicted to most likely follow
        `examples[example_idx].prompt`.

    Example
    -------
    Some story analysis::

        from llama_cpp import Llama
        from cappr import Example
        from cappr.llama_cpp.classify import predict_examples

        # Load model
        # The top of this page has instructions to download this model
        model_path = "./TinyLLama-v0.Q8_0.gguf"
        model = Llama(model_path, verbose=False)

        # Create examples
        examples = [
            Example(
                prompt="Story: I enjoyed pizza with my buddies.\\nMoral:",
                completions=("make friends", "food is yummy", "absolutely nothing"),
                prior=(2 / 5, 2 / 5, 1 / 5),
            ),
            Example(
                prompt="The child rescued the animal. The child is a",
                completions=("hero", "villain"),
            ),
        ]

        # Compute
        preds = predict_examples(examples, model)

        # the moral of the 1st story
        print(preds[0])
        # food is yummy

        # the character of the 2nd story
        print(preds[1])
        # hero

    """
    return predict_proba_examples(**locals())



"""
Helper functions to test that a `classify` module's:
  - outputs have the correct shape and form (content is NOT checked)
  - inputs are correctly checked.
"""

from __future__ import annotations
import re
from typing import Sequence

import numpy as np
import numpy.typing as npt
import pytest

from cappr import Example
from cappr.utils.classify import _ndim

from _protocol import classify_module


def _test_log_probs_conditional(
    log_probs_completions: list[list[float]] | list[list[list[float]]],
    expected_len: int,
    num_completions_per_prompt: int | None,
):
    # TODO: test that lengths of inner-most lists match # tokens in corresponding
    # completion
    if num_completions_per_prompt is None:
        # There's only one prompt and it was fed by itself, not in a sequence.
        assert len(log_probs_completions) == expected_len
    else:
        assert len(log_probs_completions) == expected_len
        # Test lengths before zipping
        assert len(log_probs_completions) == len(num_completions_per_prompt)
        for log_probs, num_completions in zip(
            log_probs_completions, num_completions_per_prompt
        ):
            assert len(log_probs) == num_completions


def log_probs_conditional(
    classify: classify_module,
    prompts: str | Sequence[str],
    completions: Sequence[str],
    *args,
    **kwargs,
):
    """
    Tests that `classify.log_probs_conditional(prompts, completions, *args, **kwargs)`
    returns a list with the right depth and length.
    """
    log_probs_completions = classify.log_probs_conditional(
        prompts, completions, *args, **kwargs
    )
    if isinstance(prompts, str):
        assert _ndim(log_probs_completions) == 2
        expected_len = len(completions)
        num_completions_per_prompt = None
    else:
        assert _ndim(log_probs_completions) == 3
        expected_len = len(prompts)
        num_completions_per_prompt = [len(completions)] * len(prompts)
    _test_log_probs_conditional(
        log_probs_completions, expected_len, num_completions_per_prompt
    )


def log_probs_conditional_examples(
    classify: classify_module,
    examples: Example | Sequence[Example],
    *args,
    **kwargs,
):
    """
    Tests that
    `classify.log_probs_conditional_examples(prompts, completions, *args, **kwargs)`
    returns a list with the right depth and length.
    """
    log_probs_completions = classify.log_probs_conditional_examples(
        examples, *args, **kwargs
    )
    if isinstance(examples, Example):
        assert _ndim(log_probs_completions) == 2
        expected_len = len(examples.completions)
        num_completions_per_prompt = None
    else:
        assert _ndim(log_probs_completions) == 3
        expected_len = len(examples)
        num_completions_per_prompt = [len(example.completions) for example in examples]
    _test_log_probs_conditional(
        log_probs_completions, expected_len, num_completions_per_prompt
    )


def predict_proba(
    classify: classify_module,
    prompts: str | Sequence[str],
    completions: Sequence[str],
    *args,
    **kwargs,
):
    """
    Tests that `classify.predict_proba(prompts, completions, *args, **kwargs)` returns a
    numpy array with the correct shape, and that each row is a probability distribution
    if `normalize=True`.
    """
    pred_probs = classify.predict_proba(prompts, completions, *args, **kwargs)
    assert isinstance(pred_probs, np.ndarray)
    if isinstance(prompts, str):
        assert pred_probs.shape == (len(completions),)
        completions_axis = 0
    else:
        assert pred_probs.shape == (len(prompts), len(completions))
        completions_axis = 1
    assert np.all(pred_probs >= 0)
    assert np.all(pred_probs <= 1)

    if pred_probs.shape[-1] == 1:
        if isinstance(prompts, str):
            p = pred_probs
        else:
            p = pred_probs[:, 0]
        # We don't normalize if there's one completion. It's almost definitely the case
        # that the predicted probability is < 1 by a decent amount.
        assert not any(np.isclose(p, 1))
    else:
        prob_sums = pred_probs.sum(axis=completions_axis)
        are_close_to_1 = np.isclose(prob_sums, 1)
        if kwargs.get("normalize", True):
            assert np.all(are_close_to_1)
        else:
            assert not np.any(are_close_to_1)

    # Test bad inputs
    if kwargs.get("prior", None) is not None:
        prior_good: list[float] = list(kwargs["prior"])

        # Wrong length
        prior_bad = prior_good + [0]
        expected_error_msg = re.escape(
            f"Expected prior to have length {len(completions)} (the number of "
            f"completions), got {len(prior_bad)}."
        )
        with pytest.raises(ValueError, match=expected_error_msg):
            _kwargs = {**kwargs, "prior": prior_bad}
            classify.predict_proba(prompts, completions, *args, **_kwargs)

        # Not a probability distr
        prior_bad = prior_good[:-1] + [0]
        expected_error_msg = "prior must sum to 1."
        with pytest.raises(ValueError, match=expected_error_msg):
            _kwargs = {**kwargs, "prior": prior_bad}
            classify.predict_proba(prompts, completions, *args, **_kwargs)

    expected_error_msg = "Setting normalize=True when there's only 1 completion"
    with pytest.raises(ValueError, match=expected_error_msg):
        _kwargs = {**kwargs, "prior": None, "normalize": True}
        classify.predict_proba(prompts, completions[:1], *args, **_kwargs)


def predict_proba_examples(
    classify: classify_module, examples: Example | Sequence[Example], *args, **kwargs
):
    """
    Tests that `classify.predict_proba_examples(examples, *args, **kwargs)` returns a
    list with the correct shape, and that each element is a probability distribution.
    """

    def test_single_example(
        pred_prob_example: npt.NDArray[np.floating], example: Example
    ):
        assert pred_prob_example.shape == (len(example.completions),)
        assert np.all(pred_prob_example >= 0)
        assert np.all(pred_prob_example <= 1)
        if len(example.completions) == 1:
            # We don't normalize if there's one completion. It's almost definitely the
            # case that the predicted probability is < 1 by a decent amount.
            assert not any(np.isclose(pred_prob_example, 1))
        else:
            prob_sum = pred_prob_example.sum()
            is_close_to_1 = np.isclose(prob_sum, 1)
            if example.normalize:
                assert is_close_to_1
            else:
                assert not is_close_to_1

    pred_probs = classify.predict_proba_examples(examples, *args, **kwargs)
    if isinstance(examples, Example):
        test_single_example(pred_probs, examples)
    else:
        assert len(pred_probs) == len(examples)
        for pred_prob_example, example in zip(pred_probs, examples):
            test_single_example(pred_prob_example, example)
    # Don't need to test bad inputs b/c they're tested in tests/test_Example.py


def predict(
    classify: classify_module,
    prompts: str | Sequence[str],
    completions: Sequence[str],
    *args,
    **kwargs,
):
    """
    Tests that `classify.predict(prompts, completions, *args, **kwargs)` returns an
    array with length `len(prompts)`, and that each element is in `completions`.
    """
    preds = classify.predict(prompts, completions, *args, **kwargs)
    # If completions is a pandas Series, then __contains__ checks the Series index, not
    # the values! Just convert to tuple.
    completions = tuple(completions)
    if isinstance(prompts, str):
        assert isinstance(preds, str)
        assert preds in completions
    else:
        assert len(preds) == len(prompts)
        assert all([pred in completions for pred in preds])

    # Test bad inputs
    with pytest.raises(ValueError, match="prompts must be non-empty."):
        classify.log_probs_conditional([], completions, *args, **kwargs)

    with pytest.raises(TypeError, match="prompts must be an ordered collection."):
        classify.log_probs_conditional(set(prompts), completions, *args, **kwargs)

    if "end_of_prompt" in kwargs:
        with pytest.raises(
            ValueError,
            match='end_of_prompt must be a whitespace " " or an empty string "".',
        ):
            _kwargs = {**kwargs, "end_of_prompt": ": "}
            classify.log_probs_conditional(prompts, completions, *args, **_kwargs)

    with pytest.raises(ValueError, match="completions must be non-empty."):
        classify.log_probs_conditional(prompts, [], *args, **kwargs)

    with pytest.raises(TypeError, match="completions must be an ordered collection."):
        classify.log_probs_conditional(prompts, set(completions), *args, **kwargs)

    with pytest.raises(TypeError, match="completions cannot be a string."):
        classify.log_probs_conditional(prompts, completions[0], *args, **kwargs)


def predict_examples(
    classify: classify_module, examples: Example | Sequence[Example], *args, **kwargs
):
    """
    Tests that `classify.predict_examples(examples, *args, **kwargs)` returns an array
    with length `len(examples)`, and that each element is one of the example's
    completions.
    """
    preds = classify.predict_examples(examples, *args, **kwargs)
    if isinstance(examples, Example):
        assert isinstance(preds, str)
        assert preds in examples.completions
    else:
        assert len(preds) == len(examples)
        for pred, example in zip(preds, examples):
            assert pred in example.completions

    # Test bad inputs
    with pytest.raises(ValueError, match="examples must be non-empty."):
        classify.log_probs_conditional_examples([], *args, **kwargs)

    is_many_examples = not isinstance(examples, Example)

    def is_hashable(objects):
        try:
            set(objects)
        except TypeError:
            return False
        else:
            return True

    if is_many_examples and is_hashable(examples):
        # if example.completions is a list, or a prior is an array, then the example
        # isn't hashable => set(examples) isn't possible in the first place
        with pytest.raises(TypeError, match="examples must be an ordered collection."):
            classify.log_probs_conditional_examples(set(examples), *args, **kwargs)



"""
Base classes which are parametrized with a set of test cases which every `classify`
module must pass.
"""

from __future__ import annotations
from typing import Sequence

import pandas as pd
import pytest

from cappr import Example

from _protocol import classify_module
import _test_form
import _test_content


class _BaseTest:
    def __setattr__(self, attr, value):
        raise AttributeError(
            f"Tried to set {attr} to {value}. But this class is frozen."
        )

    @property
    def module(self) -> classify_module:
        """
        `classify` module which needs to be tested.
        """
        raise NotImplementedError

    @property
    def module_correct(self) -> classify_module:
        """
        `classify` module which is correct, and serves as a reference implementation.
        Have it return `None` if there's no way to determine whether an implementation
        is correct. Usually that's the case if you're integrating an API endpoint like
        OpenAI.
        """
        raise NotImplementedError

    def _test(self, function: str, *args, **kwargs):
        """
        Format of all tests: test `function`'s outputs for form/structure, and then test
        their content if there exists a reference implementation to test against.
        """
        test_form = getattr(_test_form, function)
        test_form(self.module, *args, **kwargs)
        if self.module_correct is None:
            return
        test_form(self.module_correct, *args, **kwargs)
        test_content = getattr(_test_content, function)
        test_content(self.module_correct, self.module, *args, **kwargs)


@pytest.mark.parametrize(
    "prompts",
    (
        ####################### Single tokens for easy debugging #######################
        ["a b c", "c"],
        ############################### Test single-input ##############################
        "prompts single string",
        ############# Test a different type of sequence w/ corrupt indices #############
        pd.Series(["prompts is", "a", "Series"], index=[9, 0, 9]),
    ),
)
@pytest.mark.parametrize(
    "completions",
    (
        ####################### Single tokens for easy debugging #######################
        ["d e f g", "1 2", "O"],
        ######################## Test Single-token optimization ########################
        ["d", "e", "f"],
        ############# Test a different type of sequence w/ corrupt indices #############
        pd.Series(["completions is", "a", "Series"], index=[8, 10, 8]),
    ),
)
class TestPromptsCompletions(_BaseTest):
    """
    Test non-`_examples` functions with a basic set of combinations of `prompts` and
    `completions`.
    """

    def test_log_probs_conditional(
        self, prompts: str | Sequence[str], completions: Sequence[str], *args, **kwargs
    ):
        self._test("log_probs_conditional", prompts, completions, *args, **kwargs)

    # Fixtures for predict_proba
    # They're not predict_proba parametrizations b/c that raises an error indicating,
    # IIUC, that they're not in scope for this class' subclasses
    @pytest.fixture(scope="class", params=[None, 1])
    def prior(self, request: pytest.FixtureRequest):
        return request.param

    @pytest.fixture(scope="class", params=[True, False])
    def normalize(self, request: pytest.FixtureRequest) -> bool:
        return request.param

    @pytest.fixture(scope="class", params=[0, 1])
    def discount_completions(self, request: pytest.FixtureRequest) -> float:
        return request.param

    def test_predict_proba(
        self,
        prompts: str | Sequence[str],
        completions: Sequence[str],
        *args,
        prior,
        discount_completions: float,
        normalize: bool,
        **kwargs,
    ):
        self._test(
            "predict_proba",
            prompts,
            completions,
            *args,
            prior=prior if prior is None else [1 / len(completions)] * len(completions),
            discount_completions=discount_completions,
            normalize=normalize,
            **kwargs,
        )

    def test_predict(
        self, prompts: str | Sequence[str], completions: Sequence[str], *args, **kwargs
    ):
        self._test("predict", prompts, completions, *args, **kwargs)


@pytest.mark.parametrize(
    "examples",
    (
        ######################## Test non-constant # completions #######################
        [
            Example("a b c", ["d", "e f g"]),
            Example("chi", ["can", "ery"]),
            Example("koyaa", ["nisqatsi"], normalize=False),
        ],
        ############## Test constant # completions, non-constant # tokens ##############
        [
            Example(
                "jelly", ["fin", "is"], prior=pd.Series([0.33, 0.67], index=[1, 1])
            ),
            Example("a great", ["thing.", "shout"]),
            Example("out to", ["open", "source, yo."]),
        ],
        ############################ Test singleton example ############################
        Example("lonesome", ["singleton", "example"]),
        ################### Test singleton example single completion ###################
        Example("lonely", ["loner"], normalize=False),
    ),
)
class TestExamples(_BaseTest):
    """
    Test `_examples` functions with a basic set of :class:`cappr.Example`(s).
    """

    def test_log_probs_conditional_examples(
        self, examples: Example | Sequence[Example], *args, **kwargs
    ):
        self._test("log_probs_conditional_examples", examples, *args, **kwargs)

    def test_predict_proba_examples(
        self, examples: Example | Sequence[Example], *args, **kwargs
    ):
        self._test("predict_proba_examples", examples, *args, **kwargs)

    def test_predict_examples(
        self, examples: Example | Sequence[Example], *args, **kwargs
    ):
        self._test("predict_examples", examples, *args, **kwargs)


@pytest.mark.parametrize("prompt_prefix", ("a b c",))
@pytest.mark.parametrize("prompts", (["d", "d e"],))  # TODO: add single prompt
@pytest.mark.parametrize(
    "completions",
    (
        ["e f", "f g h i j"],  # multiple tokens
        ["e", "f"],  # single tokens
    ),
)
class TestCache(_BaseTest):
    def _test_log_probs_conditional(
        self,
        log_probs_completions_from_cached: list[list[list[float]]],
        prompt_prefix: str,
        prompts: list[str],
        completions: list[str],
        *model_args,
        **kwargs,
    ):
        _test_form._test_log_probs_conditional(
            log_probs_completions_from_cached,
            expected_len=len(prompts),
            num_completions_per_prompt=[len(completions)] * len(prompts),
        )
        prompts_full = [prompt_prefix + " " + prompt for prompt in prompts]
        log_probs_completions_wo_cache = self.module_correct.log_probs_conditional(
            prompts_full, completions, *model_args, **kwargs
        )
        _test_content._test_log_probs_conditional(
            log_probs_completions_from_cached,
            log_probs_completions_wo_cache,
            is_single_input=False,
        )

    def _test_log_probs_conditional_examples(
        self,
        log_probs_completions_from_cached: list[list[list[float]]],
        prompt_prefix: str,
        examples: list[Example],
        *model_args,
        **kwargs,
    ):
        _test_form._test_log_probs_conditional(
            log_probs_completions_from_cached,
            expected_len=len(examples),
            num_completions_per_prompt=[
                len(example.completions) for example in examples
            ],
        )
        examples_full = [
            Example(prompt_prefix + " " + example.prompt, example.completions)
            for example in examples
        ]
        log_probs_completions_wo_cache = (
            self.module_correct.log_probs_conditional_examples(
                examples_full, *model_args, **kwargs
            )
        )
        _test_content._test_log_probs_conditional(
            log_probs_completions_from_cached,
            log_probs_completions_wo_cache,
            is_single_input=False,
        )

    def test_cache(
        self,
        prompt_prefix: str,
        prompts: list[str],
        completions: list[str],
        *model_args,
        **kwargs,
    ):
        examples = [Example(prompt, completions) for prompt in prompts]
        with self.module.cache(*model_args, prompt_prefix) as cached:
            log_probs_completions = self.module.log_probs_conditional(
                prompts, completions, cached, **kwargs
            )
            log_probs_completions_ex = self.module.log_probs_conditional_examples(
                examples, cached, **kwargs
            )
        self._test_log_probs_conditional(
            log_probs_completions,
            prompt_prefix,
            prompts,
            completions,
            *model_args,
            **kwargs,
        )
        self._test_log_probs_conditional_examples(
            log_probs_completions_ex, prompt_prefix, examples, *model_args, **kwargs
        )

    def test_cache_model(
        self,
        prompt_prefix: str,
        prompts: list[str],
        completions: list[str],
        *model_args,
        **kwargs,
    ):
        examples = [Example(prompt, completions) for prompt in prompts]
        cached = self.module.cache_model(*model_args, prompt_prefix)
        log_probs_completions = self.module.log_probs_conditional(
            prompts, completions, cached, **kwargs
        )
        log_probs_completions_ex = self.module.log_probs_conditional_examples(
            examples, cached, **kwargs
        )
        self._test_log_probs_conditional(
            log_probs_completions,
            prompt_prefix,
            prompts,
            completions,
            *model_args,
            **kwargs,
        )
        self._test_log_probs_conditional_examples(
            log_probs_completions_ex, prompt_prefix, examples, *model_args, **kwargs
        )



"""
Transform completion token log-probabilites into a probability distribution over
completions
"""

from __future__ import annotations
from contextlib import contextmanager
from functools import wraps
from inspect import getmodule
from typing import Callable, Sequence
import warnings

import numpy as np
import numpy.typing as npt

from cappr.utils import _check


@contextmanager
def _setattr(obj, name: str, value):
    """
    In this context, `obj.name` is set to `value`.
    """
    hasattr_ = hasattr(obj, name)
    if hasattr_:
        value_prev = getattr(obj, name)
    try:
        if hasattr_:
            setattr(obj, name, value)
        yield
    finally:
        if hasattr_:
            setattr(obj, name, value_prev)


########################################################################################
############################## Aggregate log probabilities #############################
########################################################################################


def _agg_log_probs_vectorized(
    log_probs: Sequence[Sequence[Sequence[float]]],
    func: Callable[[Sequence[Sequence[float]]], Sequence[float]] = np.mean,
) -> npt.NDArray[np.floating]:
    """
    Aggregate using a vectorized numpy function `func`.
    """
    num_completions_per_prompt = [
        len(log_probs_completions) for log_probs_completions in log_probs
    ]
    num_completions_per_prompt_set = set(num_completions_per_prompt)
    if not len(num_completions_per_prompt_set) == 1:
        raise ValueError(
            "log_probs does not have a constant number of completions, i.e., there are "
            "indices i, j such that len(log_probs[i]) != len(log_probs[j])."
        )
    # At this point, we've verified that the number of completions is constant. (numpy
    # will automatically check if the number of tokens is constant later. For
    # demonstration purposes, say that the number of tokens is also constant.)
    # Say, e.g., we have 2 completions, ['a b', 'c d e'], and 2 prompts
    # Then log_probs looks like:
    # [ [ [a1, b1],      (token log-probs for completion 1 | prompt 1)
    #     [c1, d1, e1]], (token log-probs for completion 2 | prompt 1)
    #   [ [a2, b2],      (token log-probs for completion 1 | prompt 2)
    #     [c2, d2, e2]]  (token log-probs for completion 2 | prompt 2)
    # ]
    # We can re-shape this "jagged" list as a list of (non-jagged) arrays:
    # [ array([[a1, b1]],
    #         [[a2, b2]]),
    #   array([[c1, d1, e1]],
    #         [[c2, d2, e2]])
    # ]
    num_completions_per_prompt = list(num_completions_per_prompt_set)[0]
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "error",
            category=np.VisibleDeprecationWarning,
            message="Creating an ndarray from ragged nested sequences",
        )
        # If there are a non-constant # of tokens, vectorization is not possible. In
        # older versions of numpy, this error is just a warning. So catch the warning
        # and raise it as an error
        try:
            array_list = [
                np.array(
                    [
                        log_probs_completions[completion_idx]
                        for log_probs_completions in log_probs
                    ]
                )
                for completion_idx in range(num_completions_per_prompt)
            ]
        except (np.VisibleDeprecationWarning, ValueError) as exception:
            raise ValueError(
                "log_probs has a constant # of completions, but there are a "
                "non-constant # of tokens. Vectorization is not possible."
            ) from exception
    # Now apply the vectorized function to each array in the list
    likelihoods: npt.NDArray[np.floating] = np.exp(
        [func(array, axis=1) for array in array_list]
    )
    # likelihoods looks like:
    # array([[likelihood_a1b1,   likelihood_a2b2  ],
    #        [likelihood_c1d1e1, likelihood_c2d2e2]])
    # Transpose it to satisfy likelihoods[i][j] = exp(func(log_probs[i][j]))
    return likelihoods.T


def _agg_log_probs(
    log_probs: Sequence[Sequence[Sequence[float]]],
    func: Callable[[Sequence[float]], float] = np.mean,
) -> list[list[float]]:
    """
    Aggregate using a slow, nested list comprehension.
    """
    return [
        [
            np.exp(func(log_probs_completion))
            for log_probs_completion in log_probs_completions
        ]
        for log_probs_completions in log_probs
    ]


def _ndim(sequence) -> int:
    """
    Like `array.ndim` or `len(np.shape(sequence))` but `sequence` can be more jagged. It
    can't be nested inconsistently though, since we'll just check the first element.
    """

    def _is_sliceable(object) -> bool:
        try:
            object[0:0]
        except (TypeError, IndexError, ValueError):
            return False
        else:
            return True

    ndim = 0
    while _is_sliceable(sequence):
        ndim += 1
        sequence = sequence[0]
    return ndim


def agg_log_probs(
    log_probs: Sequence[Sequence[float]] | Sequence[Sequence[Sequence[float]]],
    func: Callable[[Sequence[float]], float] = np.mean,
) -> npt.NDArray[np.floating] | list[float] | list[list[float]]:
    """
    Aggregate token log-probabilities along the last dimension into probabilities.

    Parameters
    ----------
    log_probs : Sequence[Sequence[float]] | Sequence[Sequence[Sequence[float]]]
        nested sequences where token log-probabilities are in the last dimension. A 2-D
        sequence corresponds to inputting a single prompt string or
        :class:`cappr.Example` object. A 3-D sequence corresponds to inputting multiple
        prompt strings or :class:`cappr.Example` objects
    func : Callable[[Sequence[float]], float], optional
        function which aggregates a sequence of token log-probabilities into a single
        log-probability. If the function is vectorized, it must take an ``axis``
        argument. By default, `numpy.mean`

    Returns
    -------
    probs: npt.NDArray[np.floating] | list[float] | list[list[float]]
        If `log_probs` is 2-D, then `probs` is an array or list of probabilities where::

            probs[j] = exp(func(log_probs[j]))

        If `log_probs` is 3-D, then `probs` is an array or a list of list of
        probabilities where::

            probs[i][j] = exp(func(log_probs[i][j]))

    Raises
    ------
    ValueError
        if `log_probs` is not 2-D or 3-D
    """
    # 1. Determine the dimensionality of log_probs. The aggregation computation assumes
    #    a 3-D input, so we'll wrap it if it's 2-D
    ndim = _ndim(log_probs)
    if ndim not in {2, 3}:
        raise ValueError(f"log_probs must be 2-D or 3-D. Got {ndim} dimensions.")
    log_probs = [log_probs] if ndim == 2 else log_probs

    # 2. Run the aggregation computation, vectorizing if possible
    try:
        likelihoods = _agg_log_probs_vectorized(log_probs, func)
    except ValueError:  # log_probs is doubly jagged
        likelihoods = _agg_log_probs(log_probs, func)

    # 3. If we wrapped it, unwrap it for user convenience
    return likelihoods[0] if ndim == 2 else likelihoods


def posterior_prob(
    likelihoods: npt.ArrayLike[float],
    axis: int = -1,
    prior: Sequence[float] | None = None,
    normalize: bool | Sequence[bool] = True,
    check_prior: bool = True,
) -> npt.NDArray[np.floating]:
    """
    Compute posterior probabilities from likelihoods and a prior.

    Parameters
    ----------
    likelihoods : npt.ArrayLike[float]
        1-D or 2-D array of probabilities of data given a hypothesis
    axis : int, optional
        the axis/dimension along which the probability distribution is defined. By
        default, the last axis
    prior : Sequence[float] | None, optional
        a probability distribution over the `axis` of `likelihoods`. By default, a
        uniform prior is used
    normalize : bool | Sequence[bool], optional
        whether or not to return a normalized probability distribtution for each row, by
        default True (normalize all rows)
    check_prior : bool, optional
        whether or not to check that the `prior` is indeed a probability distribution,
        by default True

    Returns
    -------
    posterior_probs : npt.NDArray[np.floating]
        array of probabilities of a hypothesis given data. Its shape is the same as
        `likelihood.shape`

    Raises
    ------
    ValueError
        if `normalize` is a sequence whose length is different than that of
        `likelihoods`
    """
    # Input checks and preprocessing
    if not isinstance(likelihoods, np.ndarray):
        likelihoods = np.array(likelihoods)  # it cannot be jagged/inhomogenous
    if not isinstance(normalize, (Sequence, np.ndarray)):
        # For code simplicity, just repeat it
        # If likelihoods is 1-D, there's only a single probability distr to normalize
        num_repeats = 1 if len(likelihoods.shape) == 1 else likelihoods.shape[0]
        normalize = [normalize] * num_repeats
    elif len(normalize) != len(likelihoods):
        raise ValueError(
            "If normalize is a Sequence, it must have the same length as likelihoods. "
            f"Got {len(normalize)}, {len(likelihoods)}."
        )
    normalize = np.array(normalize, dtype=bool)
    if check_prior:
        prior = _check.prior(prior, expected_length=likelihoods.shape[axis])

    # Apply Bayes' rule, w/ optional normalization per row
    if prior is None:
        posteriors_unnorm = likelihoods
    else:
        posteriors_unnorm = likelihoods * prior
    marginals = posteriors_unnorm.sum(axis=axis, keepdims=True)
    marginals[~normalize] = 1  # denominator of 1 means no normalization
    return posteriors_unnorm / marginals


########################################################################################
############################ Decorators for classify modules ###########################
########################################################################################


def _wrap_call_unwrap(
    type_indicating_singleness: type, func: Callable, input, *args, **kwargs
):
    """
    Handles single inputs for a `func` call which only takes multiple inputs and returns
    multiple outputs.
    """
    is_single_input = isinstance(input, type_indicating_singleness)
    input = [input] if is_single_input else input
    output = func(input, *args, **kwargs)
    return output[0] if is_single_input else output


def _token_logprobs(token_log_probs):
    """
    Decorator which does input checking, and allows for `texts` to be a single string
    for a `token_log_probs` function.
    """

    @wraps(token_log_probs)
    def wrapper(
        texts: str | Sequence[str], *args, **kwargs
    ) -> list[float] | list[list[float]]:
        _check.nonempty_and_ordered(texts, variable_name="texts")
        if not isinstance(texts, str):
            texts = list(texts)  # 0-index
        _check.end_of_prompt(kwargs.get("end_of_prompt", " "))
        return _wrap_call_unwrap(str, token_log_probs, texts, *args, **kwargs)

    return wrapper


def _log_probs_conditional(log_probs_conditional):
    """
    Decorator which does input checking, and allows for `prompts` to be a single string
    for a `log_probs_conditional` function.
    """

    @wraps(log_probs_conditional)
    def wrapper(
        prompts: str | Sequence[str], completions: Sequence[str], *args, **kwargs
    ) -> list[list[float]] | list[list[list[float]]]:
        _check.nonempty_and_ordered(prompts, variable_name="prompts")
        _check.completions(completions)
        _check.end_of_prompt(kwargs.get("end_of_prompt", " "))
        return _wrap_call_unwrap(
            str, log_probs_conditional, prompts, completions, *args, **kwargs
        )

    return wrapper


def _log_probs_conditional_examples(log_probs_conditional_examples):
    """
    Decorator which does input checking, and allows for `examples` to be a single
    `Example` for a `log_probs_conditional_examples` function.
    """

    from cappr import Example

    @wraps(log_probs_conditional_examples)
    def wrapper(
        examples: Example | Sequence[Example], *args, **kwargs
    ) -> list[list[float]] | list[list[list[float]]]:
        if not isinstance(examples, Example):
            _check.nonempty_and_ordered(examples, variable_name="examples")
        return _wrap_call_unwrap(
            Example, log_probs_conditional_examples, examples, *args, **kwargs
        )

    return wrapper


def _discount(
    token_logprobs_func,
    completions: Sequence[str],
    log_probs_completions: list[list[float]] | list[list[list[float]]],
    is_single_input: bool,
    *model_args,  # rest are kwarg-only, and must come from the og function's kwargs
    discount_completions: float,
    log_marg_probs_completions: Sequence[Sequence[float]] | None = None,
    **kwargs,
) -> list[npt.NDArray[np.floating]] | list[list[npt.NDArray[np.floating]]]:
    """
    Highly experimental feature: discount completion given prompt probabilities by
    completion probabilities. Useful when particular completions are getting
    over-predicted. Currently isn't used by `_examples` functions.
    """
    # log Pr(completion token i | completion tokens :i) for each completion
    if log_marg_probs_completions is None:
        log_marg_probs_completions = token_logprobs_func(
            completions, *model_args, **kwargs
        )
    for x in log_marg_probs_completions:
        if x[0] is None:
            x[0] = 0  # no discount for the first token

    # pre-multiply by the discount amount
    log_marg_probs_completions_discounted = [
        discount_completions * np.array(log_marginal_probs_completion)
        for log_marginal_probs_completion in log_marg_probs_completions
    ]

    # add discount
    if is_single_input:
        return [
            np.array(log_probs_completions[completion_idx])
            + log_marg_probs_completions_discounted[completion_idx]
            for completion_idx in range(len(completions))
        ]
    return [
        [
            np.array(log_probs_prompt_completions[completion_idx])
            + log_marg_probs_completions_discounted[completion_idx]
            for completion_idx in range(len(completions))
        ]
        for log_probs_prompt_completions in log_probs_completions
    ]


def _predict_proba(log_probs_conditional):
    """
    Decorator which converts a `log_probs_condtional` function call into a
    `predict_proba` call.
    """

    @wraps(log_probs_conditional)
    def wrapper(
        prompts: str | Sequence[str], completions: Sequence[str], *args, **kwargs
    ) -> npt.NDArray[np.floating]:
        # 1. Check inputs before making expensive model calls

        # Check the prior
        prior = kwargs.get("prior", None)
        prior = _check.prior(prior, expected_length=len(completions))

        # Check normalization
        normalize = kwargs.get("normalize", True)
        _check.normalize(normalize, completions)

        # Check inputs for discount feature
        discount_completions = kwargs.get("discount_completions", 0)
        log_marg_probs_completions = kwargs.get("log_marg_probs_completions", None)
        if not discount_completions and (log_marg_probs_completions is not None):
            raise TypeError(
                "log_marg_probs_completions is set, but they will not be used "
                "because discount_completions was not set."
            )

        # 2. Do the expensive model calls
        log_probs_completions = log_probs_conditional(
            prompts, completions, *args, **kwargs
        )

        is_single_input = isinstance(prompts, str)

        # Maybe apply discount
        if discount_completions:
            log_probs_completions = _discount(
                getattr(getmodule(log_probs_conditional), "token_logprobs"),
                completions,
                log_probs_completions,
                is_single_input,
                *args,
                **kwargs,
            )

        # 3. Aggregate probs
        likelihoods = agg_log_probs(log_probs_completions)
        axis = 0 if is_single_input else 1
        return posterior_prob(likelihoods, axis=axis, prior=prior, normalize=normalize)

    return wrapper


def _predict_proba_examples(log_probs_conditional_examples):
    """
    Decorator which converts a `log_probs_conditional_examples` function call into a
    `predict_proba_examples` call.
    """

    from cappr import Example

    @wraps(log_probs_conditional_examples)
    def wrapper(
        examples: Example | Sequence[Example], *args, **kwargs
    ) -> npt.NDArray[np.floating] | list[npt.NDArray[np.floating]]:
        log_probs_completions = log_probs_conditional_examples(
            examples, *args, **kwargs
        )
        likelihoods = agg_log_probs(log_probs_completions)
        if isinstance(examples, Example):
            return posterior_prob(
                likelihoods,
                axis=0,
                prior=examples.prior,
                normalize=examples.normalize,
                check_prior=False,  # already checked during example construction
            )

        # Determine whether vectorization is possible for the posterior probability calc
        num_completions_per_prompt = [len(example.completions) for example in examples]
        normalize = [example.normalize for example in examples]
        num_completions_per_prompt_set = set(num_completions_per_prompt)
        if len(num_completions_per_prompt_set) != 1:
            # Can't be easily vectorized :-(
            return [
                posterior_prob(
                    likelihoods_ex,
                    axis=0,
                    prior=example.prior,
                    normalize=normalize_ex,
                    check_prior=False,  # already checked during example construction
                )
                for likelihoods_ex, example, normalize_ex in zip(
                    likelihoods, examples, normalize
                )
            ]
        # Vectorize!
        if all([example.prior is None for example in examples]):
            prior = None
        else:
            # For coding simplicity, just supply a prior which is non-None *everywhere*
            num_completions_per_prompt = list(num_completions_per_prompt_set)[0]
            uniform_prior = [
                1 / num_completions_per_prompt
            ] * num_completions_per_prompt
            prior = np.array(
                [
                    uniform_prior if example.prior is None else example.prior
                    for example in examples
                ]
            )  # same shape as likelihoods: (len(examples), num_completions_per_prompt)
        return posterior_prob(
            likelihoods,
            axis=1,
            prior=prior,
            normalize=normalize,
            check_prior=False,  # already checked during example construction
        )

    return wrapper


def _predict(predict_proba_func):
    """
    Decorator which converts a `predict_proba` function call into a `predict` call.
    """

    @wraps(predict_proba_func)
    def wrapper(
        prompts: str | Sequence[str], completions: Sequence[str], *args, **kwargs
    ) -> str | list[str]:
        if len(completions) == 1:
            raise ValueError(
                "completions only has one completion. predict will trivially return "
                "back this completion. Perhaps you meant to call predict_proba with "
                "normalize=False instead of predict."
            )
        pred_probs: npt.NDArray = predict_proba_func(
            prompts, completions, *args, **kwargs
        )
        # We need completions to support 0-indexed __getitem__
        completions = list(completions)
        num_dimensions = pred_probs.ndim
        if isinstance(prompts, str):
            # User convenience: prompts was a single string, so pred_probs is 1-D
            assert num_dimensions == 1
            return completions[pred_probs.argmax()]
        assert num_dimensions == 2
        pred_class_idxs = pred_probs.argmax(axis=1)
        return [completions[pred_class_idx] for pred_class_idx in pred_class_idxs]

    return wrapper


def _predict_examples(predict_proba_examples_func):
    """
    Decorator which converts a `predict_proba_examples` function call into a
    `predict_examples` call.
    """

    from cappr import Example

    @wraps(predict_proba_examples_func)
    def wrapper(
        examples: Example | Sequence[Example], *args, **kwargs
    ) -> str | list[str]:
        pred_probs: npt.NDArray[np.floating] | list[npt.NDArray[np.floating]] = (
            predict_proba_examples_func(examples, *args, **kwargs)
        )
        if isinstance(examples, Example):
            # User convenience: examples is a singleton
            assert pred_probs.ndim == 1  # double check
            pred_class_idx = pred_probs.argmax()
            return examples.completions[pred_class_idx]
        try:
            # If it's an array, we can call .argmax on the whole thing, which is faster
            pred_class_idxs = pred_probs.argmax(axis=1)
        except (
            AttributeError,  # no argmax attr
            TypeError,  # no axis kwarg
        ):
            pred_class_idxs = [
                example_pred_probs.argmax() for example_pred_probs in pred_probs
            ]
        return [
            example.completions[pred_class_idx]
            for example, pred_class_idx in zip(examples, pred_class_idxs)
        ]

    return wrapper



from __future__ import annotations
from contextlib import _GeneratorContextManager
from typing import Protocol, Sequence

import numpy as np
import numpy.typing as npt

from cappr import Example


Model = type("Model", (object,), {})


class classify_module(Protocol):
    """
    Protocol for all `classify` modules.
    """

    def token_logprobs(
        self, texts: str | Sequence[str], *args, **kwargs
    ) -> list[list[float]]:
        """
        For each text, log Pr(token[i] | tokens[:i]) for each token
        """

    def cache(
        self, model: Model, prefix: str, *args, **kwargs
    ) -> _GeneratorContextManager[Model]:
        """
        Optional: context manager which caches the `model` with `prefix`
        """

    def cache_model(self, model: Model, prefix: str, *args, **kwargs) -> Model:
        """
        Optional: returns a model cached with `prefix`
        """

    def log_probs_conditional(
        self, prompts: str | Sequence[str], completions: Sequence[str], *args, **kwargs
    ) -> list[list[float]] | list[list[list[float]]]:
        """
        list[i][j][k] = log Pr(
            completions[j][token k] | prompts[i] + completions[j][tokens :k]
        )
        """

    def log_probs_conditional_examples(
        self, examples: Example | Sequence[Example], *args, **kwargs
    ) -> list[list[float]] | list[list[list[float]]]:
        """
        list[i][j][k] = log Pr(
            examples[i].completions[j][token k]
            | examples[i].prompt + examples[i].completions[j][tokens :k]
        )
        """

    def predict_proba(
        self, prompts: str | Sequence[str], completions: Sequence[str], *args, **kwargs
    ) -> npt.NDArray[np.floating]:
        """
        array[i, j] = Pr(completions[j] | prompts[i])
        """

    def predict_proba_examples(
        self, examples: Example | Sequence[Example], *args, **kwargs
    ) -> npt.NDArray[np.floating] | list[npt.NDArray[np.floating]]:
        """
        list[i][j] = Pr(examples[i].completions[j] | examples[i].prompt)
        """

    def predict(
        self, prompts: str | Sequence[str], completions: Sequence[str], *args, **kwargs
    ) -> str | list[str]:
        """
        list[i] = argmax_j Pr(completions[j] | prompts[i])
        """

    def predict_examples(
        self, examples: Example | Sequence[Example], *args, **kwargs
    ) -> str | list[str]:
        """
        list[i] = argmax_j Pr(examples[i].completions[j] | examples[i].prompt)
        """



"""
Helper functions to test that outputs of `classify` modules have the correct content,
assuming a reference/correct implementation exists.
"""

from __future__ import annotations
from typing import Sequence

import torch

from cappr import Example
from _protocol import classify_module


atol = 1e-4  # TODO: fixture?


def _as_tensor(data: Sequence) -> torch.Tensor:
    # Note: throughout this module, we'll cast to torch float16 tensors to be consistent
    # with the way numerical closeness is defined for model-dependent outputs.
    dtype = torch.float16
    if isinstance(data, torch.Tensor):
        return data.to(dtype)
    return torch.tensor(data).to(dtype)


def token_logprobs(
    log_probs_texts_observed: Sequence[Sequence[float]],
    log_probs_texts_from_unbatched: Sequence[Sequence[Sequence[float]]],
    input_ids_from_unbatched: Sequence[Sequence[int]],
):
    # TODO: separate form and content checks, test number of tokens

    # The first logprob of every text must be None b/c no CausalLM estimates Pr(token)
    for log_probs_text_observed in log_probs_texts_observed:
        assert log_probs_text_observed[0] is None

    # Slice out log probs for the final expected result
    log_probs_texts_expected = []
    for _text_input_ids, _text_log_probs in zip(
        input_ids_from_unbatched, log_probs_texts_from_unbatched
    ):
        log_probs_texts_expected.append(
            [None]  # for the first token, no CausalLM estimates Pr(token)
            + [  # this token's data contains the next token's log-probability
                _text_log_probs[i, _text_input_ids[i + 1]]
                for i in range(0, len(_text_input_ids) - 1)
            ]
        )

    # Every log prob is correct, and sizes are correct
    assert len(log_probs_texts_observed) == len(log_probs_texts_expected)
    for log_probs_text_observed, log_probs_text_expected in zip(
        log_probs_texts_observed, log_probs_texts_expected
    ):
        assert len(log_probs_text_observed) == len(log_probs_text_expected)
        # skip the first token b/c its log prob is always None
        for log_prob_token_observed, log_prob_token_expected in zip(
            log_probs_text_observed[1:], log_probs_text_expected[1:]
        ):
            assert torch.isclose(
                _as_tensor(log_prob_token_observed),
                _as_tensor(log_prob_token_expected),
                atol=atol,
            )


def _test_log_probs_conditional(
    log_probs_completions1: list[list[float]] | list[list[list[float]]],
    log_probs_completions2: list[list[float]] | list[list[list[float]]],
    is_single_input: bool,
):
    def test_single_input(log_probs1: list[list[float]], log_probs2: list[list[float]]):
        assert len(log_probs1) == len(log_probs2)
        for log_probs_tokens1, log_probs_tokens2 in zip(log_probs1, log_probs2):
            assert torch.allclose(
                _as_tensor(log_probs_tokens1), _as_tensor(log_probs_tokens2), atol=atol
            )

    if is_single_input:
        test_single_input(log_probs_completions1, log_probs_completions2)
    else:
        assert len(log_probs_completions1) == len(log_probs_completions2)
        for log_probs1, log_probs2 in zip(
            log_probs_completions1, log_probs_completions2
        ):
            test_single_input(log_probs1, log_probs2)


def log_probs_conditional(
    classify1: classify_module,
    classify2: classify_module,
    prompts: str | Sequence[str],
    completions: Sequence[str],
    *args,
    **kwargs,
):
    """
    Tests that the conditional token log-probabilities are numerically close.
    """
    log_probs_completions1 = classify1.log_probs_conditional(
        prompts, completions, *args, **kwargs
    )
    log_probs_completions2 = classify2.log_probs_conditional(
        prompts, completions, *args, **kwargs
    )
    _test_log_probs_conditional(
        log_probs_completions1,
        log_probs_completions2,
        is_single_input=isinstance(prompts, str),
    )


def log_probs_conditional_examples(
    classify1: classify_module,
    classify2: classify_module,
    examples: Example | Sequence[Example],
    *args,
    **kwargs,
):
    """
    Tests that the conditional token log-probabilities are numerically close.
    """
    log_probs_completions1 = classify1.log_probs_conditional_examples(
        examples, *args, **kwargs
    )
    log_probs_completions2 = classify2.log_probs_conditional_examples(
        examples, *args, **kwargs
    )
    _test_log_probs_conditional(
        log_probs_completions1,
        log_probs_completions2,
        is_single_input=isinstance(examples, Example),
    )


def predict_proba(
    classify1: classify_module,
    classify2: classify_module,
    prompts: str | Sequence[str],
    completions: Sequence[str],
    *args,
    **kwargs,
):
    """
    Tests that predicted probabilities are numerically close.
    """
    pred_probs1 = classify1.predict_proba(prompts, completions, *args, **kwargs)
    pred_probs2 = classify2.predict_proba(prompts, completions, *args, **kwargs)
    assert torch.allclose(_as_tensor(pred_probs1), _as_tensor(pred_probs2), atol=atol)


def predict_proba_examples(
    classify1: classify_module,
    classify2: classify_module,
    examples: Example | Sequence[Example],
    *args,
    **kwargs,
):
    """
    Tests that predicted probabilities are numerically close.
    """
    pred_probs1 = classify1.predict_proba_examples(examples, *args, **kwargs)
    pred_probs2 = classify2.predict_proba_examples(examples, *args, **kwargs)
    for pred_probs1_ex, pred_probs2_ex in zip(pred_probs1, pred_probs2):
        assert torch.allclose(
            _as_tensor(pred_probs1_ex), _as_tensor(pred_probs2_ex), atol=atol
        )


def predict(
    classify1: classify_module,
    classify2: classify_module,
    prompts: str | Sequence[str],
    completions: Sequence[str],
    *args,
    **kwargs,
):
    """
    Tests that predictions are identical.
    """
    preds1 = classify1.predict(prompts, completions, *args, **kwargs)
    preds2 = classify2.predict(prompts, completions, *args, **kwargs)
    assert preds1 == preds2


def predict_examples(
    classify1: classify_module,
    classify2: classify_module,
    examples: Example | Sequence[Example],
    *args,
    **kwargs,
):
    """
    Tests that predictions are identical.
    """
    preds1 = classify1.predict_examples(examples, *args, **kwargs)
    preds2 = classify2.predict_examples(examples, *args, **kwargs)
    assert preds1 == preds2



"""
Helpers functions which are compatible with the Python OpenAI API (before and after
v1.0.0)
"""

from __future__ import annotations
from contextlib import contextmanager, nullcontext
from dataclasses import dataclass
import logging
import os
import time
from typing import Any, Callable, Literal, Sequence
import warnings

import openai

try:
    from openai import OpenAI
except ImportError:
    OpenAI = type("OpenAI", (object,), {})
    _IS_OPENAI_AFTER_V1 = False
    _RETRY_ERRORS = (
        openai.error.ServiceUnavailableError,
        openai.error.RateLimitError,
    )
else:
    _IS_OPENAI_AFTER_V1 = True
    _RETRY_ERRORS = (openai.InternalServerError, openai.RateLimitError)
import tiktoken

from cappr.utils import _batch, _check


logger = logging.getLogger(__name__)


Model = Literal[  # Literal so that docs and IDEs easily parse it
    "babbage-002",
    "davinci-002",
]
"""
These are /v1/completions models where `echo=True, logprobs=1` can be passed. On
October 5, 2023, OpenAI deprecated this combination of arguments for the
``gpt-3.5-turbo-instruct`` model. So it cannot supported by CAPPr.
"""


@dataclass(frozen=True)
class _DollarCostPer1kTokens:
    """
    Represents the cost of processing prompt and completion text.

    Parameters
    ----------
    prompt : float | None
        dollar cost for inputting 1k prompt tokens. `None` means cost is unknown.
    completion: float | None
        dollar cost for generating 1k completion tokens. `None` means cost is unknown.
    """

    prompt: float | None
    completion: float | None


# https://openai.com/api/pricing/
# TODO: figure out how to get this automatically from openai, if possible
_MODEL_TO_COST_PER_1K_TOKENS = {
    # /v1/chat/completions
    "gpt-4": _DollarCostPer1kTokens(prompt=0.03, completion=0.06),
    # /v1/completions
    "babbage-002": _DollarCostPer1kTokens(prompt=0.0004, completion=0.0004),
    "davinci-002": _DollarCostPer1kTokens(prompt=0.002, completion=0.002),
    # Deprecated echo=True, logprobs=1 on 10/05/2023
    "gpt-3.5-turbo-instruct": _DollarCostPer1kTokens(prompt=0.0005, completion=0.0015),
}


class _UserCanceled(Exception):
    pass


def _openai_api_call_is_ok(
    texts: list[str],
    model: Model,
    max_tokens: int = 0,
    cost_per_1k_tokens_prompt: float | None = None,
    cost_per_1k_tokens_completion: float | None = None,
) -> tuple[int, int, int | str]:
    """
    After displaying the cost (usually an upper bound) of hitting the OpenAI API
    text completion endpoint, prompt the user to manually input ``y`` or ``n`` to
    indicate whether the program can proceed.

    Parameters
    ----------
    texts : list[str]
        texts or prompts inputted to the `model` OpenAI API call
    model : Model
        name of the OpenAI API text completion model
    max_tokens : int, optional
        maximum number of tokens to generate, by default 0
    cost_per_1k_tokens_prompt : float | None, optional
        OpenAI API dollar cost for processing 1k prompt tokens. If unset,
        `cappr.openai.api._MODEL_TO_COST_PER_1K_TOKENS[model]["prompt"]` is used. If
        it's still unknown, the cost will be displayed as `unknown`. By default, the
        cost is unknown
    cost_per_1k_tokens_completion : float | None, optional
        OpenAI API dollar cost for processing 1k completion tokens. If unset,
        `cappr.openai.api._MODEL_TO_COST_PER_1K_TOKENS[model]["completion"]` is used. If
        it's still unknown, the cost will be displayed as `unknown`. By default, the
        cost is unknown

    Returns
    -------
    tuple[int, int, int | str]
        - number of prompt tokens, i.e., the number of tokens in `texts`
        - upper bound on number of completion tokens, i.e., `len(texts) * max_tokens`
        - if the `model`'s costs are known, the dollar cost of having it process
        `texts`. Else, a string which is a URL to OpenAI's pricing page.

    Raises
    ------
    _UserCanceled
        if the user inputs ``n`` when prompted to give the go-ahead
    """
    texts = list(texts)
    try:
        tokenizer = tiktoken.encoding_for_model(model)
    except KeyError:  # that's fine, we just need an approximation
        tokenizer = tiktoken.get_encoding("gpt2")

    _dollar_cost_unknown = _DollarCostPer1kTokens(prompt=None, completion=None)

    # prompts
    num_tokens_prompts = sum(len(tokens) for tokens in tokenizer.encode_batch(texts))
    cost_per_1k_tokens_prompt = (
        cost_per_1k_tokens_prompt
        or _MODEL_TO_COST_PER_1K_TOKENS.get(model, _dollar_cost_unknown).prompt
    )
    if cost_per_1k_tokens_prompt is None:
        cost = None
    else:
        cost = num_tokens_prompts * cost_per_1k_tokens_prompt / 1_000

    # completions
    num_tokens_completions = len(texts) * max_tokens  # upper bound
    cost_per_1k_tokens_completion = (
        cost_per_1k_tokens_completion
        or _MODEL_TO_COST_PER_1K_TOKENS.get(model, _dollar_cost_unknown).completion
    )
    if cost is not None and cost_per_1k_tokens_completion is not None:
        cost += num_tokens_completions * cost_per_1k_tokens_completion / 1_000
        cost = round(cost, 2)
    else:
        cost = "unknown (see https://openai.com/api/pricing/)"

    num_tokens_total = num_tokens_prompts + num_tokens_completions
    output = None
    while output not in {"y", "n"}:
        output = input(
            f"This API call will cost about ${cost} (≤{num_tokens_total:_} tokens). "
            "Proceed? (y/n): "
        )
    if output == "n":
        raise _UserCanceled("No API requests will be submitted.")
    return num_tokens_prompts, num_tokens_completions, cost  # for testing purposes


def openai_method_retry(
    openai_method: Callable,
    max_num_tries: int = 5,
    sleep_sec: float = 10,
    retry_errors: tuple = _RETRY_ERRORS,
    **openai_method_kwargs,
):
    """
    Wrapper around OpenAI API calls which automatically retries and sleeps if requests
    fail. Logs at level INFO when requests fail, and level ERROR if an exception is
    raised.

    Parameters
    ----------
    openai_method : Callable
        a function or method whose inputs are `openai_method_kwargs`
    max_num_tries : int, optional
        maximum number of times to retry the request before raising the exception, by
        default 5
    sleep_sec : float, optional
        number of seconds to sleep before re-submitting the request, by default 10
    retry_errors : tuple[Exception], optional
        if one of these exceptions is raised by the request, then retry, else the
        exception is immediately raised. By default, these are the
        ServiceUnavailableError/InternalServerError and RateLimitError.

    Returns
    -------
    Any
        `openai_method(**openai_method_kwargs)`

    Raises
    ------
    Exception
        if `max_num_tries` is exceeded or an exception not in `retry_errors` is raised
    """
    num_tries = 0
    while num_tries < max_num_tries:
        try:
            return openai_method(**openai_method_kwargs)
        except retry_errors as e:
            num_tries += 1
            logger.info(f"openai error: {e}")
            logger.info(f"Try {num_tries}. Sleeping for {sleep_sec} sec.")
            exception = e  # allow it to be referenced later
            time.sleep(sleep_sec)
    logger.error(f"Max retries exceeded. openai error: {exception}")
    raise exception


@contextmanager
def _set_openai_api_key(api_key: str | None = None):
    """
    In this context, the OpenAI module attribute ``openai.api_key`` is set to whatever
    is closest in scope.
    """
    api_key_from_module = openai.api_key
    try:
        # Priority is kinda in increasing order of scope
        openai.api_key = api_key or api_key_from_module or os.getenv("OPENAI_API_KEY")
        # The openai module will raise a good error (when a request is submitted) if
        # this module attr is None in the end
        yield
    finally:
        openai.api_key = api_key_from_module


def _to_dict(response) -> dict[str, Any]:
    if not hasattr(response, "model_dump"):
        # It's a dict b/c we're on openai < v1.0.0
        return response
    else:
        # It's a pydantic model b/c we're on openai >= v1.0.0
        dump_nested_dicts = getattr(response, "model_dump")
        # When the model is dumped, it gives a Pydantic serializer warning for each
        # token: Expected `int` but got `float`
        # https://github.com/openai/openai-python/issues/744
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                message="Pydantic serializer warnings",
            )
            response = dump_nested_dicts()
        return response


def gpt_complete(
    texts: Sequence[str],
    model: Model,
    client: OpenAI | None = None,
    show_progress_bar: bool | None = None,
    progress_bar_desc: str = "log-probs",
    ask_if_ok: bool = False,
    api_key: str | None = None,
    max_tokens: int = 0,
    **openai_completion_kwargs,
) -> list[dict[str, Any]]:
    """
    Wrapper around the OpenAI text completion endpoint which automatically batches
    texts for greater efficiency, retries requests that fail, and displays a progress
    bar.

    OpenAI API text completion reference:
    https://platform.openai.com/docs/api-reference/completions

    Warning
    -------
    By default, no tokens will be generated/sampled.

    Parameters
    ----------
    texts : Sequence[str]
        these are passed as the `prompt` argument in a text completion request
    model : Model
        which text completion model to use
    client : OpenAI | None, optional
        an OpenAI client object. By default, the global client instance is used
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 texts
    progress_bar_desc: str, optional
        description of the progress bar if shown, by default ``'log-probs'``
    ask_if_ok : bool, optional
        whether or not to prompt you to manually give the go-ahead to run this function,
        after notifying you of the approximate cost of the OpenAI API calls. By default,
        False
    api_key : str | None, optional
        your OpenAI API key. By default, it's set to the OpenAI's module attribute
        ``openai.api_key``, or the environment variable ``OPENAI_API_KEY``
    max_tokens : int, optional
        maximum number of tokens to generate, by default 0
    **openai_completion_kwargs
        other arguments passed to the text completion endpoint, e.g., `logprobs=1`

    Returns
    -------
    list[dict[str, Any]]
        list with the same length as `texts`. Each element is the ``choices`` mapping
    """
    _check.ordered(texts, variable_name="texts")
    if not _IS_OPENAI_AFTER_V1:
        openai_method = openai.Completion.create
    else:
        openai_method = (
            openai.completions.create if client is None else client.completions.create
        )
    with _set_openai_api_key(api_key) if client is None else nullcontext():
        _batch_size = 20  # max that the API can currently handle
        # Passing in a string will silently but majorly fail. Handle it
        texts = [texts] if isinstance(texts, str) else texts
        if ask_if_ok:
            _ = _openai_api_call_is_ok(texts, model, max_tokens=max_tokens)
        choices: list[dict[str, Any]] = []
        with _batch.ProgressBar(
            total=len(texts),
            show_progress_bar=show_progress_bar,
            desc=progress_bar_desc,
        ) as progress_bar:
            for texts_batch in _batch.constant(texts, _batch_size):
                response = openai_method_retry(
                    openai_method,
                    prompt=texts_batch,
                    model=model,
                    max_tokens=max_tokens,
                    **openai_completion_kwargs,
                )
                response = _to_dict(response)
                choices.extend(response["choices"])
                progress_bar.update(len(texts_batch))
        return choices


def gpt_chat_complete(
    texts: Sequence[str],
    model: str = "gpt-3.5-turbo",
    client: OpenAI | None = None,
    show_progress_bar: bool | None = None,
    ask_if_ok: bool = False,
    api_key: str | None = None,
    system_msg: str = "You are an assistant which classifies text.",
    max_tokens: int = 5,
    temperature: float = 0,
    **openai_chat_kwargs,
) -> list[dict[str, Any]]:
    """
    Wrapper around the OpenAI chat completion endpoint which sends `texts` 1-by-1 as
    individual chat messages in a chat. It retries requests that fail and displays a
    progress bar.

    OpenAI API chat completion reference:
    https://platform.openai.com/docs/api-reference/chat

    Warning
    -------
    By default, the `system_msg` asks ChatGPT to perform text classification. And the
    `temperature` is set to 0.

    Parameters
    ----------
    texts : Sequence[str]
        texts which are passed in one by one immediately after the system content as
        ``{"role": "user", "content": text}``
    model : str, optional
        one of the chat model names, by default "gpt-3.5-turbo"
    client : OpenAI | None, optional
        an OpenAI client object. By default, the global client instance is used
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 texts
    ask_if_ok : bool, optional
        whether or not to prompt you to manually give the go-ahead to run this function,
        after notifying you of the approximate cost of the OpenAI API calls. By default,
        False
    api_key : str | None, optional
        your OpenAI API key. By default, it's set to the OpenAI's module attribute
        ``openai.api_key``, or the environment variable ``OPENAI_API_KEY``
    system_msg : str, optional
        text which is passed in 1-by-1 immediately before every piece of user content in
        `texts` as ``{"role": "system", "content": system_msg}``. By default ``"You are
        an assistant which classifies text."``
    max_tokens : int, optional
        maximum number of tokens to generate, by default 5
    temperature : float, optional
        probability re-scaler to apply before sampling, by default 0
    **openai_chat_kwargs
        other arguments passed to the chat completion endpoint

    Returns
    -------
    list[dict[str, Any]]
        list with the same length as `texts`. Each element is the ``choices`` mapping
    """
    _check.ordered(texts, variable_name="texts")
    if not _IS_OPENAI_AFTER_V1:
        openai_method = openai.ChatCompletion.create
    else:
        openai_method = (
            openai.chat.completions.create
            if client is None
            else client.chat.completions.create
        )
    # TODO: batch, if possible
    with _set_openai_api_key(api_key) if client is None else nullcontext():
        # Passing in a string will silently but majorly fail. Handle it
        texts = [texts] if isinstance(texts, str) else texts
        if ask_if_ok:
            _ = _openai_api_call_is_ok(texts, model, max_tokens=max_tokens)
        choices: list[dict[str, Any]] = []
        for text in _batch.ProgressBar(
            texts, show_progress_bar=show_progress_bar, desc="Completing chats"
        ):
            messages = [
                {"role": "system", "content": system_msg},
                {"role": "user", "content": text},
            ]
            response = openai_method_retry(
                openai_method,
                model=model,
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                **openai_chat_kwargs,
            )
            response = _to_dict(response)
            choices.extend(response["choices"])
        return choices



"""
Perform prompt-completion classification using a model which can be loaded via

- ``transformers.AutoModelForCausalLM.from_pretrained`` or
- ``auto_gptq.AutoGPTQForCausalLM.from_quantized``.

You probably just want the :func:`predict` or :func:`predict_examples` functions :-)

In the implementation, attention block keys and values for prompts are automatically
cached and shared across completions.
"""

from __future__ import annotations
from contextlib import contextmanager, nullcontext
from dataclasses import dataclass
from typing import cast, Literal, Mapping, Sequence, Tuple

import numpy as np
import numpy.typing as npt
import torch
from transformers import PreTrainedTokenizerBase
from transformers.modeling_outputs import CausalLMOutput, CausalLMOutputWithPast

from cappr.utils import _batch, _check, classify
from cappr import Example
from cappr import huggingface as hf
from cappr.huggingface._utils import BatchEncodingPT, ModelForCausalLM


@classify._token_logprobs
@_batch.flatten
@_batch.batchify(batchable_arg="texts", progress_bar_desc="log-probs")
def token_logprobs(
    texts: str | Sequence[str],
    model_and_tokenizer: tuple[ModelForCausalLM, PreTrainedTokenizerBase],
    end_of_prompt: Literal[" ", ""] = " ",
    show_progress_bar: bool | None = None,
    add_bos: bool = False,
    batch_size: int = 16,
    **kwargs,
) -> list[float] | list[list[float]]:
    """
    For each text, compute each token's log-probability conditional on all previous
    tokens in the text.

    Parameters
    ----------
    texts : str | Sequence[str]
        input text(s)
    model_and_tokenizer : tuple[ModelForCausalLM, PreTrainedTokenizerBase]
        a model and its tokenizer
    end_of_prompt : Literal[' ', ''], optional
        This string gets added to the beginning of each text. It's important to set this
        if you're using the discount feature. Otherwise, set it to "". By default " "
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 texts
    add_bos : bool, optional
        whether or not to add a beginning-of-sentence token to each text in `texts` if
        the tokenizer has a beginning-of-sentence token, by default False
    batch_size : int, optional
        the maximum number of `texts` that the model will process in parallel, by
        default 16

    Returns
    -------
    log_probs : list[float] | list[list[float]]

        If `texts` is a string, then a 1-D list is returned: `log_probs[token_idx]` is
        the log-probability of the token at `token_idx` of `texts` conditional on all
        previous tokens in `texts`.

        If `texts` is a sequence of strings, then a 2-D list is returned:
        `log_probs[text_idx][token_idx]` is the log-probability of the token at
        `token_idx` of `texts[text_idx]` conditional on all previous tokens in
        `texts[text_idx]`.

    Warning
    -------
    Set `end_of_prompt="", add_bos=True` unless you're using the discount feature.

    Note
    ----
    For each text, the first token's log-probability is always ``None`` because no
    autoregressive LM directly estimates the marginal probability of a token.

    Raises
    ------
    TypeError
        if `texts` is not a sequence
    ValueError
        if `texts` is empty
    """
    model, tokenizer = model_and_tokenizer
    if not hf._utils.does_tokenizer_need_prepended_space(tokenizer):
        end_of_prompt = ""
    texts = [end_of_prompt + text for text in texts]

    with hf._utils.dont_add_bos_token(tokenizer) if not add_bos else nullcontext():
        logits, encodings = hf._utils.logits_texts(
            texts, (model, tokenizer), drop_bos_token=not add_bos
        )
    log_probs_texts = hf._utils.logits_to_log_probs(
        logits=logits,
        input_ids=encodings["input_ids"],
        input_ids_start_idx=1,  # this token's log-prob is in the prev token's logit
        logits_end_idx=-1,
    )

    # Remove pad token logprobs
    num_non_pad_tokens = encodings["attention_mask"].sum(dim=1)
    log_probs = []
    first_token_log_prob = [None]
    for log_probs_text, n in zip(log_probs_texts, num_non_pad_tokens):
        # Slice off the right side b/c the tokenizer was set up to do right-padding
        log_probs.append(first_token_log_prob + log_probs_text[: (n - 1)].tolist())
    return log_probs


########################################################################################
########################## Attention past_key_values utilities #########################
########################################################################################


_PastKeyValues = Tuple[Tuple[torch.Tensor, torch.Tensor]]
"""
The `past_key_values` input to a HuggingFace `transformers` model's forward pass. It's a
2-D tuple of 4-D tensors.

Index items:

(
    # attention blocks = 12 for gpt2,

    2 = (attention keys, attention values), (or 6 for LongLlama)

    batch size = # input texts,

    # heads = 12 for gpt2,

    max # tokens in batch,

    key/value hidden dimension = 64 for gpt2
)
"""


# past_key_values must be a tuple for model calls. So we have to do slightly costly
# transfers from Python to CUDA. I don't think there's anything we can do about that


def _past_key_values_tuple_to_tensor(past_key_values: _PastKeyValues) -> torch.Tensor:
    if past_key_values is None:
        raise TypeError(
            "past_key_values is None. Can your model be configured to output it? If "
            "not, please use cappr.huggingface.classify_no_cache"
        )  # pragma: no cover
    return torch.stack([torch.stack(block) for block in past_key_values], dim=0)


def _past_key_values_tensor_to_tuple(past_key_values: torch.Tensor) -> _PastKeyValues:
    return tuple(
        tuple(block[i] for i in range(len(block))) for block in past_key_values
    )


def _past_key_values_get(
    past_key_values: _PastKeyValues,
    batch_idxs: Sequence[int],
) -> _PastKeyValues:
    return _past_key_values_tensor_to_tuple(
        _past_key_values_tuple_to_tensor(past_key_values)[:, :, batch_idxs, ...]
    )


########################################################################################
############################# KV caching + batch inference #############################
########################################################################################


@dataclass
class _CAPPr:
    model: ModelForCausalLM
    logits_all: bool = True
    _past: tuple[BatchEncodingPT, CausalLMOutputWithPast] | None = None
    batch_idxs: torch.Tensor | None = None
    update_cache: bool = False
    _is_cache_cleared: bool = False

    @property
    def past(self):
        return self._past

    @past.setter
    def past(self, new_past: tuple[BatchEncodingPT, CausalLMOutputWithPast] | None):
        if new_past is None:
            self._is_cache_cleared = True
        self._past = new_past


class _CacheClearedError(Exception):
    """Raise to prevent a user from using a cached model outside of the context"""


class _ModelWithCache:
    def __init__(
        self,
        model: ModelForCausalLM,
        encodings_to_cache: BatchEncodingPT,
        past: tuple[BatchEncodingPT, CausalLMOutputWithPast] | None = None,
        logits_all: bool = True,
    ):
        self._cappr = _CAPPr(model, logits_all, past)
        """
        Contains data which controls the cache
        """
        # This data is in one place to minimize pollution of the inputted model's
        # namespace. This object should be treated like a ModelForCausalLM by the user
        self._cappr.update_cache = True
        _ = self.forward(**encodings_to_cache)
        del _
        self._cappr.update_cache = False

    def forward(
        self, input_ids: torch.Tensor, attention_mask: torch.Tensor
    ) -> CausalLMOutputWithPast:
        if self._cappr._is_cache_cleared:
            raise _CacheClearedError(
                "This model is no longer usable. It was used in a context where "
                "clear_cache_on_exit=True. If you instead meant to retain the cache "
                "for future use, then use: `cached_model_and_tokenizer = "
                "cache_model(model_and_tokenizer, prefixes)`"
            )

        encodings = {"input_ids": input_ids, "attention_mask": attention_mask}

        if self._cappr.past is None:
            with hf._utils.set_up_model(self._cappr.model):
                out: CausalLMOutputWithPast = self._cappr.model(**encodings)
            self._cappr.past = encodings, out
            return out

        encodings_past, out_past = self._cappr.past

        if self._cappr.batch_idxs is None:
            # Attempt to broadcast the past with the present
            if encodings_past["input_ids"].shape[0] == 1 and input_ids.shape[0] > 1:
                # Assume the past of present input[i] is past[0] (the only past)
                batch_idxs = torch.zeros(
                    input_ids.shape[0], device=self._cappr.model.device, dtype=torch.int
                )
            elif encodings_past["input_ids"].shape[0] == input_ids.shape[0]:
                # Assume the past of present input[i] is past[i]
                batch_idxs = torch.arange(
                    input_ids.shape[0], device=self._cappr.model.device
                )
            else:
                raise ValueError(
                    f'Could not broadcast {encodings_past["input_ids"].shape[0]} '
                    f"cached inputs with {input_ids.shape[0]} new inputs."
                )
        else:
            # Externally set to a tensor. Bad design but whatever
            batch_idxs: torch.Tensor = self._cappr.batch_idxs

        if not torch.equal(
            batch_idxs,
            torch.arange(
                encodings_past["input_ids"].shape[0], device=self._cappr.model.device
            ),
        ):
            # Must extract past by converting the tuple to a tensor and back
            past_key_values = _past_key_values_get(out_past.past_key_values, batch_idxs)
        else:
            past_key_values = out_past.past_key_values

        attention_mask_past = encodings_past["attention_mask"][batch_idxs]
        attention_mask_past_cat_present = torch.cat(
            [attention_mask_past, encodings["attention_mask"]], dim=1
        )
        # Set position_ids to what they'd be had we fed prompt + completion together
        # Model implementations currently don't do this, so gotta do it manually
        # Bug: https://github.com/huggingface/transformers/issues/18104
        offsets = attention_mask_past.sum(dim=1)  # numbers of nonpad tokens in past
        position_ids_present = (
            torch.arange(input_ids.shape[1], device=self._cappr.model.device)
            + offsets[:, None]
        )

        with hf._utils.set_up_model(self._cappr.model):
            out: CausalLMOutputWithPast = self._cappr.model(
                input_ids=input_ids,
                attention_mask=attention_mask_past_cat_present,
                position_ids=position_ids_present,
                past_key_values=past_key_values,
            )

        # past_key_values is already concatenated in out.past_key_values
        if self._cappr.logits_all:
            out.logits = torch.cat([out_past.logits[batch_idxs], out.logits], dim=1)
        if self._cappr.update_cache:
            # Concatenate encodings for future model calls
            input_ids_past = encodings_past["input_ids"][batch_idxs]
            encodings = {
                "input_ids": torch.cat([input_ids_past, input_ids], dim=1),
                "attention_mask": attention_mask_past_cat_present,
            }
            self._cappr.past = encodings, out
        return out

    def __call__(
        self, input_ids: torch.Tensor, attention_mask: torch.Tensor
    ) -> CausalLMOutputWithPast:
        return self.forward(input_ids, attention_mask)

    def __getattr__(self, __name: str):
        return getattr(self._cappr.model, __name)

    def __repr__(self) -> str:
        return repr(self._cappr.model)


def cache_model(
    model_and_tokenizer: tuple[ModelForCausalLM, PreTrainedTokenizerBase],
    prefixes: str | Sequence[str],
    logits_all: bool = True,
) -> tuple[ModelForCausalLM, PreTrainedTokenizerBase]:
    """
    Caches the model so that every future computation with it starts with `prefixes`. As
    a result, computations with this model are faster.

    Use this function instead of the context manager :func:`cache` to keep the cache for
    future computations, including those outside of a context.

    Parameters
    ----------
    model_and_tokenizer : tuple[ModelForCausalLM, PreTrainedTokenizerBase]
        a model and its tokenizer
    prefixes : str | Sequence[str]
        prefix(es) for all future strings that will be processed, e.g., a string
        containing shared prompt instructions, or a string containing instructions and
        exemplars for few-shot prompting
    logits_all : bool, optional
        whether or not to have the cached model include logits for all tokens (including
        the past). By default, past token logits are included

    Returns
    -------
    tuple[ModelForCausalLM, PreTrainedTokenizerBase]
        cached model and the (unmodified) tokenizer

    Note
    ----
    If you're inputting the cached model and tokenizer to a function in this module,
    e.g., :func:`predict`, `prefixes` and future strings are assumed to be separated by
    a whitespace. Otherwise, ensure that any strings that are processed by the tokenizer
    start correctly. Furthermore, if applicable, set ``tokenizer.add_bos_token = False``
    for future computations.

    Example
    -------
    Usage with :func:`predict_proba`::

        import numpy as np
        from transformers import AutoModelForCausalLM, AutoTokenizer
        from cappr.huggingface.classify import cache_model, predict_proba

        # Load model and tokenizer
        model = AutoModelForCausalLM.from_pretrained("gpt2")
        tokenizer = AutoTokenizer.from_pretrained("gpt2")
        model_and_tokenizer = (model, tokenizer)

        # Create data
        prompt_prefix = '''Instructions: complete the sequence.
        Here are examples:
        A, B, C => D
        1, 2, 3 => 4

        Complete this sequence:'''

        prompts = ["a, b, c =>", "X, Y =>"]
        completions = ["d", "Z", "Hi"]

        # Cache
        cached_model_and_tokenizer = cache_model(
            model_and_tokenizer, prompt_prefix
        )

        # Compute
        pred_probs = predict_proba(
            prompts, completions, cached_model_and_tokenizer
        )

        # The above computation is equivalent to this one:
        prompts_full = [prompt_prefix + " " + prompt for prompt in prompts]
        pred_probs_wo_cache = predict_proba(
            prompts_full, completions, model_and_tokenizer
        )
        assert np.allclose(pred_probs, pred_probs_wo_cache, atol=1e-5)

        print(pred_probs.round(1))
        # [[1. 0. 0.]
        #  [0. 1. 0.]]
    """
    _check.nonempty_and_ordered(prefixes, variable_name="prefixes")
    prefixes = [prefixes] if isinstance(prefixes, str) else list(prefixes)
    model, tokenizer = model_and_tokenizer
    try:
        past = getattr(getattr(model, "_cappr"), "past")
    except AttributeError:
        past = None

    with hf._utils.set_up_tokenizer(tokenizer):
        with nullcontext() if past is None else hf._utils.dont_add_bos_token(tokenizer):
            encodings_prefixes: BatchEncodingPT = tokenizer(
                prefixes, return_tensors="pt", padding=True
            ).to(model.device)

    model_for_causal_lm = (
        model._cappr.model if isinstance(model, _ModelWithCache) else model
    )
    model_with_cache = _ModelWithCache(
        model_for_causal_lm,
        encodings_to_cache=encodings_prefixes,
        past=past,
        logits_all=logits_all,
    )
    return model_with_cache, tokenizer  # return tokenizer for consistent interface


@contextmanager
def cache(
    model_and_tokenizer: tuple[ModelForCausalLM, PreTrainedTokenizerBase],
    prefixes: str | Sequence[str],
    clear_cache_on_exit: bool = True,
    logits_all: bool = True,
):
    """
    In this context, every prompt processed by `model_and_tokenizer` starts with a fixed
    prefix. As a result, computations in this context are faster.

    Parameters
    ----------
    model_and_tokenizer : tuple[ModelForCausalLM, PreTrainedTokenizerBase]
        a model and its tokenizer
    prefixes : str | Sequence[str]
        prefix(es) for all strings that will be processed in this context, e.g., a
        string containing shared prompt instructions, or a string containing
        instructions and exemplars for few-shot prompting
    clear_cache_on_exit : bool, optional
        whether or not to clear the cache and render the returned model and tokenizer
        unusable when we exit the context. This is important because it saves memory,
        and makes code more explicit about the model's state. By default, True
    logits_all : bool, optional
        whether or not to have the cached model include logits for all tokens (including
        the past). By default, past token logits are included

    Note
    ----
    If you're inputting the cached model and tokenizer to a function in this module,
    e.g., :func:`predict`, `prefixes` and future strings are assumed to be separated by
    a whitespace. Otherwise, ensure that any strings that are processed by the tokenizer
    start correctly.

    Example
    -------
    Usage with :func:`predict_proba`::

        import numpy as np
        from transformers import AutoModelForCausalLM, AutoTokenizer
        from cappr.huggingface.classify import cache, predict_proba

        # Load model and tokenizer
        model = AutoModelForCausalLM.from_pretrained("gpt2")
        tokenizer = AutoTokenizer.from_pretrained("gpt2")
        model_and_tokenizer = (model, tokenizer)

        # Create data
        prompt_prefix = '''Instructions: complete the sequence.
        Here are examples:
        A, B, C => D
        1, 2, 3 => 4

        Complete this sequence:'''

        prompts = ["a, b, c =>", "X, Y =>"]
        completions = ["d", "Z", "Hi"]

        # Compute
        with cache(
            model_and_tokenizer, prompt_prefix
        ) as cached_model_and_tokenizer:
            # prompt_prefix and each prompt are separated by a whitespace
            pred_probs = predict_proba(
                prompts, completions, cached_model_and_tokenizer
            )

        # The above computation is equivalent to this one:
        prompts_full = [prompt_prefix + " " + prompt for prompt in prompts]
        pred_probs_wo_cache = predict_proba(
            prompts_full, completions, model_and_tokenizer
        )
        assert np.allclose(pred_probs, pred_probs_wo_cache, atol=1e-5)

        print(pred_probs.round(1))
        # [[1. 0. 0.]
        #  [0. 1. 0.]]

    Here's a more complicated example, which might help in explaining usage::

        import torch
        from transformers import AutoModelForCausalLM, AutoTokenizer
        from cappr.huggingface.classify import cache
        from cappr.huggingface._utils import (
            does_tokenizer_need_prepended_space,
            logits_texts,
        )

        # Load model and tokenizer
        model = AutoModelForCausalLM.from_pretrained("gpt2")
        tokenizer = AutoTokenizer.from_pretrained("gpt2")
        model_and_tokenizer = (model, tokenizer)

        # Assume that all strings will be separated by a whitespace
        delim = " "
        if not does_tokenizer_need_prepended_space(tokenizer):
            # for SentencePiece tokenizers like Llama's
            delim = ""

        logits = lambda *args, **kwargs: logits_texts(*args, **kwargs)[0]
        '''
        Returns next-token logits for each token in an inputted text.
        '''

        with cache(model_and_tokenizer, "a") as cached_a:
            with cache(cached_a, delim + "b c") as cached_a_b_c:
                with cache(cached_a_b_c, delim + "d") as cached_a_b_c_d:
                    logits1 = logits([delim + "e f"], cached_a_b_c_d)
                    logits2 = logits([delim + "x"], cached_a_b_c_d)
                logits3 = logits([delim + "1 2 3"], cached_a_b_c)
            logits4 = logits([delim + "b c d"], cached_a)

        logits_correct = lambda texts, **kwargs: logits(
            texts, model_and_tokenizer, drop_bos_token=False
        )

        atol = 1e-4
        assert torch.allclose(logits1, logits_correct(["a b c d e f"]), atol=atol)
        assert torch.allclose(logits2, logits_correct(["a b c d x"]), atol=atol)
        assert torch.allclose(logits3, logits_correct(["a b c 1 2 3"]), atol=atol)
        assert torch.allclose(logits4, logits_correct(["a b c d"]), atol=atol)
    """
    try:
        past = getattr(getattr(model_and_tokenizer[0], "_cappr"), "past")
    except AttributeError:
        past = None
    model_with_cache, tokenizer = cache_model(
        model_and_tokenizer, prefixes, logits_all=logits_all
    )
    model_with_cache = cast(_ModelWithCache, model_with_cache)
    # Now that model_with_cache has a past, we should never add a BOS token
    with hf._utils.dont_add_bos_token(tokenizer), hf._utils.set_up_tokenizer(tokenizer):
        yield model_with_cache, tokenizer

    if clear_cache_on_exit:
        # model_with_cache._cappr.past contains a lot of data—logits, past_key_values,
        # hidden_states (usually taking up GPU RAM)—that should be cleared when we exit
        # the context
        model_with_cache._cappr.past = None
    else:
        model_with_cache._cappr.past = past


########################################################################################
############################### Logits from cached model ###############################
########################################################################################


def _last_nonpad_token_logits(logits: torch.Tensor, attention_mask: torch.Tensor):
    last_nonpad_token_idxs = (
        logits.shape[1] - (attention_mask.flip(dims=[1]) == 1).max(dim=1).indices - 1
    )
    return logits.take_along_dim(last_nonpad_token_idxs[:, None, None], dim=1)


def _blessed_helper(
    model: ModelForCausalLM,
    tokenizer: PreTrainedTokenizerBase,
    prompts: Sequence[str],
    completions: Sequence[str],
    num_completions_per_prompt: int | Sequence[int],
    completions_repeats: int,
    batch_size_completions: int | None = None,
) -> tuple[torch.Tensor, BatchEncodingPT]:
    if not isinstance(num_completions_per_prompt, int):
        num_completions_per_prompt: torch.Tensor = torch.tensor(
            num_completions_per_prompt, device=model.device
        )
    batch_size_completions = batch_size_completions or len(completions)
    prompts_idxs: tuple[torch.Tensor] = (
        torch.arange(len(prompts), device=model.device)
        .repeat_interleave(num_completions_per_prompt)
        .split(batch_size_completions)
    )
    # Prepare completions data
    # Completions shouldn't start w/ a bos token <s> b/c we need to mimic sending the
    # prompt + completion together. For example, if 'a b' is the prompt and 'c' is the
    # completion, the encoding should correspond to '<s> a b c' not '<s> a b <s> c'
    with hf._utils.set_up_tokenizer(tokenizer), hf._utils.dont_add_bos_token(tokenizer):
        # This computation is repeated for constant completions, but it doesn't matter
        completions_encoding: BatchEncodingPT = tokenizer(
            completions, return_tensors="pt", padding=True
        ).to(model.device)
    completions_encoding = cast(BatchEncodingPT, completions_encoding)
    # Repeat then batch completions. The other way around corrupts the order
    completions_input_ids = torch.split(
        completions_encoding["input_ids"].repeat(completions_repeats, 1),
        batch_size_completions,
    )
    completions_attention_mask = torch.split(
        completions_encoding["attention_mask"].repeat(completions_repeats, 1),
        batch_size_completions,
    )
    num_batches = len(completions_input_ids)

    # TODO: put this in the context manager? Little weird
    if not hf._utils.does_tokenizer_need_prepended_space(tokenizer):
        start_of_prompt = ""
    else:
        in_cache_context = isinstance(model, _ModelWithCache)
        start_of_prompt = " " if in_cache_context else ""
    prompts = [start_of_prompt + prompt for prompt in prompts]

    # Run model, caching prompts
    with cache((model, tokenizer), prompts, logits_all=False) as cached_prompts:
        cached_prompts_model, _ = cached_prompts
        prompts_encodings, prompts_out = cached_prompts_model._cappr.past
        prompts_last_nonpad_token_logits = _last_nonpad_token_logits(
            prompts_out.logits, prompts_encodings["attention_mask"]
        )
        # fmt: off
        _are_completions_constant = (
            isinstance(num_completions_per_prompt, int) and
            completions_repeats == len(prompts)
        )
        # fmt: on
        if (
            completions_encoding["input_ids"].shape[1] == 1
            and _are_completions_constant
        ):
            # Single-token optimization: if every completion is a single token, we don't
            # need to repeat stuff or run the model on any of the completions data
            # Currently, this optimization is only done for constant completions, i.e.,
            # not _examples
            # Note that completions_encoding["input_ids"].shape[1] == logits.shape[1]
            return prompts_last_nonpad_token_logits, completions_encoding

        completions_logits = []
        _batch_idxs = cached_prompts_model._cappr.batch_idxs
        for batch_idx in range(num_batches):
            cached_prompts_model._cappr.batch_idxs = prompts_idxs[batch_idx]
            out: CausalLMOutput = cached_prompts_model(
                input_ids=completions_input_ids[batch_idx],
                attention_mask=completions_attention_mask[batch_idx],
            )
            # B/c logits_all=False, out.logits only has completion token logits
            completions_logits.append(out.logits)
        cached_prompts_model._cappr.batch_idxs = _batch_idxs
        offsets = prompts_encodings["attention_mask"].sum(dim=1)

    # Drop the next-token logits for the last completion token. They're not useful for
    # CAPPr. Moreover, dropping ensures completions_logits.shape[:2] ==
    # completions_encoding["input_ids"].shape, as one expects. Just keep in mind that
    # logits are shifted behind
    completions_logits = torch.cat(completions_logits, dim=0)[:, :-1, :]
    prompts_last_nonpad_token_logits = (
        prompts_last_nonpad_token_logits.repeat_interleave(
            num_completions_per_prompt, dim=0
        )
    )
    completions_logits = torch.cat(
        [prompts_last_nonpad_token_logits, completions_logits], dim=1
    )
    # You may need the offsets to be able to ignore pad tokens
    completions_encoding = {
        "input_ids": torch.cat(completions_input_ids),
        "attention_mask": torch.cat(completions_attention_mask),
        "offsets": offsets.repeat_interleave(num_completions_per_prompt, dim=0),
    }
    if getattr(tokenizer, "add_bos_token", False):
        # Drop the first <s> token after we're done encoding so that the shape is
        # consistent w/ other tokenizers
        completions_encoding["offsets"] = completions_encoding["offsets"] - 1
    return completions_logits, completions_encoding


def _logits_completions_given_prompts(
    model: ModelForCausalLM,
    tokenizer: PreTrainedTokenizerBase,
    prompts: Sequence[str],
    completions: Sequence[str],
    end_of_prompt: Literal[" ", ""] = " ",
    batch_size_completions: int | None = None,
) -> tuple[torch.Tensor, BatchEncodingPT]:
    if not hf._utils.does_tokenizer_need_prepended_space(tokenizer):
        end_of_prompt = ""
    completions = [end_of_prompt + completion for completion in completions]
    return _blessed_helper(
        model,
        tokenizer,
        prompts,
        completions,
        num_completions_per_prompt=len(completions),
        completions_repeats=len(prompts),
        batch_size_completions=batch_size_completions,
    )


def _logits_completions_given_prompts_examples(
    model: ModelForCausalLM,
    tokenizer: PreTrainedTokenizerBase,
    examples: Sequence[Example],
    batch_size_completions: int | None = None,
) -> tuple[torch.Tensor, BatchEncodingPT]:
    should_end_of_prompt_be_empty = not hf._utils.does_tokenizer_need_prepended_space(
        tokenizer
    )
    prompts = [example.prompt for example in examples]
    end_of_prompts = [
        "" if should_end_of_prompt_be_empty else example.end_of_prompt
        for example in examples
    ]
    completions = [
        end_of_prompt + completion
        for end_of_prompt, example in zip(end_of_prompts, examples)
        for completion in example.completions
    ]
    num_completions_per_prompt = [len(example.completions) for example in examples]
    return _blessed_helper(
        model,
        tokenizer,
        prompts,
        completions,
        num_completions_per_prompt=num_completions_per_prompt,
        completions_repeats=1,
        batch_size_completions=batch_size_completions,
    )


########################################################################################
################################## Logits to log-probs #################################
########################################################################################


def _logits_to_log_probs_completions(
    logits: torch.Tensor, encodings: Mapping[str, torch.Tensor], from_examples: bool
) -> list[list[float]]:
    if (not from_examples) and logits.shape[1] == 1:
        # Single-token optimization: all of the completions are always a single token
        # Slice out their tokens from the prompts' last non-pad token logits. Currently,
        # this optimization is only done for constant completions, i.e., not _examples
        completions_input_ids: torch.Tensor = (
            encodings["input_ids"]
            .repeat_interleave(logits.shape[0], dim=1)  # the number of prompts
            .T
        )
        log_probs = hf._utils.logits_to_log_probs(logits, completions_input_ids)
        # Need to reshape them to the expected shape
        return log_probs.flatten()[:, None].tolist()

    # There are some completions with multiple tokens
    log_probs = hf._utils.logits_to_log_probs(logits, encodings["input_ids"])
    last_idx_non_pad = encodings["attention_mask"].sum(dim=1)
    # i.e., # of tokens per completion
    return [
        log_probs_prompt_completion[:completion_end].tolist()
        for log_probs_prompt_completion, completion_end in zip(
            log_probs, last_idx_non_pad
        )
    ]


########################################################################################
##################################### Implementation ###################################
########################################################################################


@classify._log_probs_conditional
def log_probs_conditional(
    prompts: str | Sequence[str],
    completions: Sequence[str],
    model_and_tokenizer: tuple[ModelForCausalLM, PreTrainedTokenizerBase],
    end_of_prompt: Literal[" ", ""] = " ",
    show_progress_bar: bool | None = None,
    batch_size: int = 2,
    batch_size_completions: int | None = None,
    **kwargs,
) -> list[list[float]] | list[list[list[float]]]:
    """
    Log-probabilities of each completion token conditional on each prompt and previous
    completion tokens.

    Parameters
    ----------
    prompts : str | Sequence[str]
        string(s), where, e.g., each contains the text you want to classify
    completions : Sequence[str]
        strings, where, e.g., each one is the name of a class which could come after a
        prompt
    model_and_tokenizer : tuple[ModelForCausalLM, PreTrainedTokenizerBase]
        a model and its tokenizer
    end_of_prompt : Literal[' ', ''], optional
        whitespace or empty string to join prompt and completion, by default whitespace
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 prompts
    batch_size : int, optional
        the maximum number of `prompts` that the model will process in parallel, by
        default 2
    batch_size_completions : int, optional
        the maximum number of `completions` that the model will process in parallel. By
        default, all completions are processed in parallel

    Returns
    -------
    log_probs_completions : list[list[float]] | list[list[list[float]]]

        If `prompts` is a string, then a 2-D list is returned:
        `log_probs_completions[completion_idx][completion_token_idx]` is the
        log-probability of the completion token in `completions[completion_idx]`,
        conditional on `prompt + end_of_prompt` and previous completion tokens.

        If `prompts` is a sequence of strings, then a 3-D list is returned:
        `log_probs_completions[prompt_idx][completion_idx][completion_token_idx]` is the
        log-probability of the completion token in `completions[completion_idx]`,
        conditional on `prompts[prompt_idx] + end_of_prompt` and previous completion
        tokens.

    Note
    ----
    To efficiently aggregate `log_probs_completions`, use
    :func:`cappr.utils.classify.agg_log_probs`.

    Example
    -------
    Here we'll use single characters (which are single tokens) to more clearly
    demonstrate what this function does::

        from transformers import AutoModelForCausalLM, AutoTokenizer
        from cappr.huggingface.classify import log_probs_conditional

        # Load model and tokenizer
        model = AutoModelForCausalLM.from_pretrained("gpt2")
        tokenizer = AutoTokenizer.from_pretrained("gpt2")

        # Create data
        prompts = ["x y", "a b c"]
        completions = ["z", "d e"]

        # Compute
        log_probs_completions = log_probs_conditional(
            prompts, completions, model_and_tokenizer=(model, tokenizer)
        )

        # Outputs (rounded) next to their symbolic representation

        print(log_probs_completions[0])
        # [[-4.5],        [[log Pr(z | x, y)],
        #  [-5.6, -3.2]]   [log Pr(d | x, y),    log Pr(e | x, y, d)]]

        print(log_probs_completions[1])
        # [[-9.7],        [[log Pr(z | a, b, c)],
        #  [-0.2, -0.03]]  [log Pr(d | a, b, c), log Pr(e | a, b, c, d)]]
    """

    @_batch.flatten
    @_batch.batchify(batchable_arg="prompts", progress_bar_desc="conditional log-probs")
    def log_probs_completions_batch(
        prompts, show_progress_bar=show_progress_bar, batch_size=batch_size
    ):
        logits, encodings = _logits_completions_given_prompts(
            *model_and_tokenizer,
            prompts,
            completions,
            end_of_prompt=end_of_prompt,
            batch_size_completions=batch_size_completions,
        )
        return _logits_to_log_probs_completions(logits, encodings, from_examples=False)

    log_probs_completions = log_probs_completions_batch(prompts)
    return list(_batch.constant(log_probs_completions, size=len(completions)))


@classify._log_probs_conditional_examples
def log_probs_conditional_examples(
    examples: Example | Sequence[Example],
    model_and_tokenizer: tuple[ModelForCausalLM, PreTrainedTokenizerBase],
    show_progress_bar: bool | None = None,
    batch_size: int = 2,
    batch_size_completions: int | None = None,
) -> list[list[float]] | list[list[list[float]]]:
    """
    Log-probabilities of each completion token conditional on each prompt and previous
    completion tokens.

    Parameters
    ----------
    examples : Example | Sequence[Example]
        `Example` object(s), where each contains a prompt and its set of possible
        completions
    model_and_tokenizer : tuple[ModelForCausalLM, PreTrainedTokenizerBase]
        a model and its tokenizer
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 `examples`
    batch_size : int, optional
        the maximum number of `examples` that the model will process in parallel, by
        default 2
    batch_size_completions : int, optional
        the maximum number of `completions` that the model will process in parallel. By
        default, all completions are processed in parallel

    Returns
    -------
    log_probs_completions : list[list[float]] | list[list[list[float]]]

        If `examples` is a :class:`cappr.Example`, then a 2-D list is returned:
        `log_probs_completions[completion_idx][completion_token_idx]` is the
        log-probability of the completion token in
        `example.completions[completion_idx]`, conditional on `example.prompt +
        example.end_of_prompt` and previous completion tokens.

        If `examples` is a sequence of :class:`cappr.Example` objects, then a 3-D list
        is returned:
        `log_probs_completions[example_idx][completion_idx][completion_token_idx]` is
        the log-probability of the completion token in
        `examples[example_idx].completions[completion_idx]`, conditional on
        `examples[example_idx].prompt + examples[example_idx].end_of_prompt` and
        previous completion tokens.

    Note
    ----
    To aggregate `log_probs_completions`, use
    :func:`cappr.utils.classify.agg_log_probs`.

    Note
    ----
    The attribute :attr:`cappr.Example.prior` is unused.

    Example
    -------
    Here we'll use single characters (which are single tokens) to more clearly
    demonstrate what this function does::

        from transformers import AutoModelForCausalLM, AutoTokenizer
        from cappr import Example
        from cappr.huggingface.classify import log_probs_conditional_examples

        # Load model and tokenizer
        model = AutoModelForCausalLM.from_pretrained("gpt2")
        tokenizer = AutoTokenizer.from_pretrained("gpt2")

        # Create examples
        examples = [
            Example(prompt="x y", completions=("z", "d e")),
            Example(prompt="a b c", completions=("1 2",), normalize=False),
        ]

        # Compute
        log_probs_completions = log_probs_conditional_examples(
            examples, model_and_tokenizer=(model, tokenizer)
        )

        # Outputs (rounded) next to their symbolic representation

        print(log_probs_completions[0])  # corresponds to examples[0]
        # [[-4.5],        [[log Pr(z | x, y)],
        #  [-5.6, -3.2]]   [log Pr(d | x, y),    log Pr(e | x, y, d)]]

        print(log_probs_completions[1])  # corresponds to examples[1]
        # [[-5.0, -1.7]]  [[log Pr(1 | a, b, c)], log Pr(2 | a, b, c, 1)]]
    """
    # examples is always a Sequence[Example] b/c of the decorator
    examples = cast(Sequence[Example], examples)

    @_batch.flatten
    @_batch.batchify(
        batchable_arg="examples", progress_bar_desc="conditional log-probs"
    )
    def log_probs_completions_batch(
        examples, show_progress_bar=show_progress_bar, batch_size=batch_size
    ):
        logits, encodings = _logits_completions_given_prompts_examples(
            *model_and_tokenizer,
            examples,
            batch_size_completions=batch_size_completions,
        )
        return _logits_to_log_probs_completions(logits, encodings, from_examples=True)

    log_probs_completions = log_probs_completions_batch(examples)
    num_completions_per_prompt = [len(example.completions) for example in examples]
    return list(
        _batch.variable(log_probs_completions, sizes=num_completions_per_prompt)
    )


@classify._predict_proba
def predict_proba(
    prompts: str | Sequence[str],
    completions: Sequence[str],
    model_and_tokenizer: tuple[ModelForCausalLM, PreTrainedTokenizerBase],
    prior: Sequence[float] | None = None,
    end_of_prompt: Literal[" ", ""] = " ",
    normalize: bool = True,
    discount_completions: float = 0.0,
    log_marg_probs_completions: Sequence[Sequence[float]] | None = None,
    show_progress_bar: bool | None = None,
    batch_size: int = 2,
    batch_size_completions: int | None = None,
) -> npt.NDArray[np.floating]:
    """
    Predict probabilities of each completion coming after each prompt.

    Parameters
    ----------
    prompts : str | Sequence[str]
        string(s), where, e.g., each contains the text you want to classify
    completions : Sequence[str]
        strings, where, e.g., each one is the name of a class which could come after a
        prompt
    model_and_tokenizer : tuple[ModelForCausalLM, PreTrainedTokenizerBase]
        a model and its tokenizer
    prior : Sequence[float] | None, optional
        a probability distribution over `completions`, representing a belief about their
        likelihoods regardless of the prompt. By default, each completion in
        `completions` is assumed to be equally likely
    end_of_prompt : Literal[' ', ''], optional
        whitespace or empty string to join prompt and completion, by default whitespace
    normalize : bool, optional
        whether or not to normalize completion-after-prompt probabilities into a
        probability distribution over completions. Set this to `False` if you'd like the
        raw completion-after-prompt probability, or you're solving a multi-label
        prediction problem. By default, True
    discount_completions : float, optional
        experimental feature: set it (e.g., 1.0 may work well) if a completion is
        consistently getting too high predicted probabilities. You could instead fudge
        the `prior`, but this hyperparameter may be easier to tune than the `prior`. By
        default 0.0
    log_marg_probs_completions : Sequence[Sequence[float]] | None, optional
        experimental feature: pre-computed log probabilities of completion tokens
        conditional on previous completion tokens (not prompt tokens). Only used if `not
        discount_completions`. Pre-compute them by passing `completions`, `model`, and
        `end_of_prompt` to :func:`token_logprobs`. By default, if `not
        discount_completions`, they are (re-)computed
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 prompts
    batch_size : int, optional
        the maximum number of `prompts` that the model will process in parallel, by
        default 2
    batch_size_completions : int, optional
        the maximum number of `completions` that the model will process in parallel. By
        default, all completions are processed in parallel

    Returns
    -------
    pred_probs : npt.NDArray[np.floating]

        If `prompts` is a string, then an array with shape `len(completions),` is
        returned: `pred_probs[completion_idx]` is the model's estimate of the
        probability that `completions[completion_idx]` comes after `prompt +
        end_of_prompt`.

        If `prompts` is a sequence of strings, then an array with shape `(len(prompts),
        len(completions))` is returned: `pred_probs[prompt_idx, completion_idx]` is the
        model's estimate of the probability that `completions[completion_idx]` comes
        after `prompts[prompt_idx] + end_of_prompt`.

    Note
    ----
    In this function, the set of possible completions which could follow each prompt is
    the same for every prompt. If instead, each prompt could be followed by a
    *different* set of completions, then construct a sequence of :class:`cappr.Example`
    objects and pass them to :func:`predict_proba_examples`.

    Example
    -------
    Let's have GPT-2 (small) predict where stuff is in the kitchen. This example also
    conveys that it's not the greatest model out there::

        from transformers import AutoModelForCausalLM, AutoTokenizer
        from cappr.huggingface.classify import predict_proba

        # Load model and tokenizer
        model = AutoModelForCausalLM.from_pretrained("gpt2")
        tokenizer = AutoTokenizer.from_pretrained("gpt2")

        # Define a classification task
        prompts = ["The tacos are cooking", "Ice cream is"]
        class_names = ("on the stove", "in the freezer", "in the fridge")
        prior = (1 / 5, 2 / 5, 2 / 5)

        pred_probs = predict_proba(
            prompts,
            completions=class_names,
            model_and_tokenizer=(model, tokenizer),
            prior=prior,
        )
        pred_probs_rounded = pred_probs.round(1)  # just for cleaner output

        # predicted probability that tacos cook on the stove
        print(pred_probs_rounded[0, 0])
        # 0.4

        # predicted probability that ice cream is in the freezer
        print(pred_probs_rounded[1, 1])
        # 0.5

        # predicted probability that ice cream is in the fridge
        print(pred_probs_rounded[1, 2])
        # 0.4
    """
    return log_probs_conditional(**locals())


@classify._predict_proba_examples
def predict_proba_examples(
    examples: Example | Sequence[Example],
    model_and_tokenizer: tuple[ModelForCausalLM, PreTrainedTokenizerBase],
    show_progress_bar: bool | None = None,
    batch_size: int = 2,
    batch_size_completions: int | None = None,
) -> npt.NDArray[np.floating] | list[npt.NDArray[np.floating]]:
    """
    Predict probabilities of each completion coming after each prompt.

    Parameters
    ----------
    examples : Example | Sequence[Example]
        `Example` object(s), where each contains a prompt and its set of possible
        completions
    model_and_tokenizer : tuple[ModelForCausalLM, PreTrainedTokenizerBase]
        a model and its tokenizer
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 `examples`
    batch_size : int, optional
        the maximum number of `examples` that the model will process in parallel, by
        default 2
    batch_size_completions : int, optional
        the maximum number of `completions` that the model will process in parallel. By
        default, all completions are processed in parallel

    Returns
    -------
    pred_probs : npt.NDArray[np.floating] | list[npt.NDArray[np.floating]]

        If `examples` is an :class:`cappr.Example`, then an array with shape
        `(len(example.completions),)` is returned: `pred_probs[completion_idx]` is the
        model's estimate of the probability that `example.completions[completion_idx]`
        comes after `example.prompt + example.end_of_prompt`.

        If `examples` is a sequence of :class:`cappr.Example` objects, then a list with
        length `len(examples)` is returned: `pred_probs[example_idx][completion_idx]` is
        the model's estimate of the probability that
        `examples[example_idx].completions[completion_idx]` comes after
        `examples[example_idx].prompt + examples[example_idx].end_of_prompt`. If the
        number of completions per example is a constant `k`, then an array with shape
        `(len(examples), k)` is returned instead of a list of 1-D arrays.

    Example
    -------
    GPT-2 (small) doing media trivia::

        from transformers import AutoModelForCausalLM, AutoTokenizer
        from cappr import Example
        from cappr.huggingface.classify import predict_proba_examples

        # Load model and tokenizer
        model = AutoModelForCausalLM.from_pretrained("gpt2")
        tokenizer = AutoTokenizer.from_pretrained("gpt2")

        # Create examples
        examples = [
            Example(
                prompt="Jodie Foster played",
                completions=("Clarice Starling", "Trinity in The Matrix"),
            ),
            Example(
                prompt="Batman, from Batman: The Animated Series, was played by",
                completions=("Pete Holmes", "Kevin Conroy", "Spongebob!"),
                prior=(1 / 3, 2 / 3, 0),
            ),
        ]

        pred_probs = predict_proba_examples(
            examples, model_and_tokenizer=(model, tokenizer)
        )

        # predicted probability that Jodie Foster played Clarice Starling, not Trinity
        print(pred_probs[0][0].round(2))
        # 0.7

        # predicted probability that Batman was played by Kevin Conroy
        print(pred_probs[1][1].round(2))
        # 0.97
    """
    return log_probs_conditional_examples(**locals())


@classify._predict
def predict(
    prompts: str | Sequence[str],
    completions: Sequence[str],
    model_and_tokenizer: tuple[ModelForCausalLM, PreTrainedTokenizerBase],
    prior: Sequence[float] | None = None,
    end_of_prompt: Literal[" ", ""] = " ",
    discount_completions: float = 0.0,
    log_marg_probs_completions: Sequence[Sequence[float]] | None = None,
    show_progress_bar: bool | None = None,
    batch_size: int = 2,
    batch_size_completions: int | None = None,
) -> str | list[str]:
    """
    Predict which completion is most likely to follow each prompt.

    Parameters
    ----------
    prompts : str | Sequence[str]
        string(s), where, e.g., each contains the text you want to classify
    completions : Sequence[str]
        strings, where, e.g., each one is the name of a class which could come after a
        prompt
    model_and_tokenizer : tuple[ModelForCausalLM, PreTrainedTokenizerBase]
        a model and its tokenizer
    prior : Sequence[float] | None, optional
        a probability distribution over `completions`, representing a belief about their
        likelihoods regardless of the prompt. By default, each completion in
        `completions` is assumed to be equally likely
    end_of_prompt : Literal[' ', ''], optional
        whitespace or empty string to join prompt and completion, by default whitespace
    discount_completions : float, optional
        experimental feature: set it to >0.0 (e.g., 1.0 may work well) if a completion
        is consistently getting over-predicted. You could instead fudge the `prior`, but
        this hyperparameter may be easier to tune than the `prior`. By default 0.0
    log_marg_probs_completions : Sequence[Sequence[float]] | None, optional
        experimental feature: pre-computed log probabilities of completion tokens
        conditional on previous completion tokens (not prompt tokens). Only used if `not
        discount_completions`. Pre-compute them by passing `completions`, `model`, and
        `end_of_prompt` to :func:`token_logprobs`. By default, if `not
        discount_completions`, they are (re-)computed
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 prompts
    batch_size : int, optional
        the maximum number of `prompts` that the model will process in parallel, by
        default 2
    batch_size_completions : int, optional
        the maximum number of `completions` that the model will process in parallel. By
        default, all completions are processed in parallel

    Returns
    -------
    preds : str | list[str]

        If `prompts` is a string, then the completion from `completions` which is
        predicted to most likely follow `prompt + end_of_prompt` is returned.

        If `prompts` is a sequence of strings, then a list with length `len(prompts)` is
        returned. `preds[prompt_idx]` is the completion in `completions` which is
        predicted to follow `prompts[prompt_idx] + end_of_prompt`.

    Note
    ----
    In this function, the set of possible completions which could follow each prompt is
    the same for every prompt. If instead, each prompt could be followed by a
    *different* set of completions, then construct a sequence of :class:`cappr.Example`
    objects and pass them to :func:`predict_examples`.

    Example
    -------
    Let's have GPT-2 (small) predict where stuff is in the kitchen::

        from transformers import AutoModelForCausalLM, AutoTokenizer
        from cappr.huggingface.classify import predict

        # Load model and tokenizer
        model = AutoModelForCausalLM.from_pretrained("gpt2")
        tokenizer = AutoTokenizer.from_pretrained("gpt2")

        # Define a classification task
        prompts = ["The tacos are cooking", "Ice cream is"]
        class_names = ("on the stove", "in the freezer", "in the fridge")
        prior = (1 / 5, 2 / 5, 2 / 5)

        preds = predict(
            prompts,
            completions=class_names,
            model_and_tokenizer=(model, tokenizer),
            prior=prior,
        )
        print(preds)
        # ['on the stove',
        #  'in the freezer']
    """
    return predict_proba(**locals())


@classify._predict_examples
def predict_examples(
    examples: Example | Sequence[Example],
    model_and_tokenizer: tuple[ModelForCausalLM, PreTrainedTokenizerBase],
    show_progress_bar: bool | None = None,
    batch_size: int = 2,
    batch_size_completions: int | None = None,
) -> str | list[str]:
    """
    Predict which completion is most likely to follow each prompt.

    Parameters
    ----------
    examples : Example | Sequence[Example]
        `Example` object(s), where each contains a prompt and its set of possible
        completions
    model_and_tokenizer : tuple[ModelForCausalLM, PreTrainedTokenizerBase]
        a model and its tokenizer
    show_progress_bar : bool | None, optional
        whether or not to show a progress bar. By default, it will be shown only if
        there are at least 5 `examples`
    batch_size : int, optional
        the maximum number of `examples` that the model will process in parallel, by
        default 2
    batch_size_completions : int, optional
        the maximum number of `completions` that the model will process in parallel. By
        default, all completions are processed in parallel

    Returns
    -------
    preds : str | list[str]

        If `examples` is an :class:`cappr.Example`, then the completion from
        `example.completions` which is predicted to most likely follow `example.prompt +
        example.end_of_prompt` is returned.

        If `examples` is a sequence of :class:`cappr.Example` objects, then a list with
        length `len(examples)` is returned: `preds[example_idx]` is the completion in
        `examples[example_idx].completions` which is predicted to most likely follow
        `examples[example_idx].prompt + examples[example_idx].end_of_prompt`.

    Example
    -------
    GPT-2 (small) doing media trivia::

        from transformers import AutoModelForCausalLM, AutoTokenizer
        from cappr import Example
        from cappr.huggingface.classify import predict_examples

        # Load model and tokenizer
        model = AutoModelForCausalLM.from_pretrained("gpt2")
        tokenizer = AutoTokenizer.from_pretrained("gpt2")

        # Create examples
        examples = [
            Example(
                prompt="Jodie Foster played",
                completions=("Clarice Starling", "Trinity in The Matrix"),
            ),
            Example(
                prompt="Batman, from Batman: The Animated Series, was played by",
                completions=("Pete Holmes", "Kevin Conroy", "Spongebob!"),
                prior=(1 / 3, 2 / 3, 0),
            ),
        ]

        preds = predict_examples(
            examples, model_and_tokenizer=(model, tokenizer)
        )
        print(preds)
        # ['Clarice Starling',
        #  'Kevin Conroy']
    """
    return predict_proba_examples(**locals())



"""
Unit tests `cappr.Example` input checks.
"""

from __future__ import annotations
import re
from typing import Any, Sequence

import numpy as np
import pandas as pd
import pytest

from cappr import Example


def modify_kwargs(kwargs: dict[str, Any], **kwargs_new) -> dict[str, Any]:
    return {**kwargs, **kwargs_new}


@pytest.mark.parametrize("prompt", ("hi",))
@pytest.mark.parametrize("completions", (["completion won", "completion tew"],))
@pytest.mark.parametrize("end_of_prompt", (" ",))
@pytest.mark.parametrize(
    "prior", (None, [1 / 3, 2 / 3], pd.Series([0.5, 0.5], index=[3, 3]))
)
@pytest.mark.parametrize("normalize", (True, False))
def test___post_init__(
    prompt: str,
    completions: list,
    end_of_prompt: str,
    prior: Sequence[float],
    normalize: bool,
):
    kwargs_valid = dict(
        prompt=prompt,
        completions=completions,
        end_of_prompt=end_of_prompt,
        prior=prior,
        normalize=normalize,
    )
    # Valid Examples
    Example(**kwargs_valid)
    Example(
        **modify_kwargs(
            kwargs_valid,
            completions=pd.Series(
                completions,
                index=np.random.choice(len(completions), size=len(completions)),
            ),
        )
    )

    # Invalid Examples

    # prompt - empty
    with pytest.raises(ValueError, match="prompt must be non-empty."):
        Example(**modify_kwargs(kwargs_valid, prompt=""))
    # prompt - not a string
    with pytest.raises(TypeError, match="prompt must be a string."):
        Example(**modify_kwargs(kwargs_valid, prompt=[prompt]))

    # end_of_prompt - non-" "/""
    _msg = 'end_of_prompt must be a whitespace " " or empty string "".'
    with pytest.raises(TypeError, match=_msg):
        Example(**modify_kwargs(kwargs_valid, end_of_prompt=None))
    with pytest.raises(ValueError, match=_msg):
        Example(**modify_kwargs(kwargs_valid, end_of_prompt=": "))

    # completions - empty
    with pytest.raises(ValueError, match="completions must be non-empty."):
        Example(**modify_kwargs(kwargs_valid, completions=[]))
    # completions - non-ordered
    with pytest.raises(TypeError, match="completions must be an ordered collection."):
        Example(**modify_kwargs(kwargs_valid, completions=set(completions)))
    # completions - string
    with pytest.raises(TypeError, match="completions cannot be a string."):
        Example(**modify_kwargs(kwargs_valid, completions=completions[0]))
    # completions - empty strings (all)
    _completions_all_empty = [""] * len(completions)
    with pytest.raises(ValueError, match="All completions are empty."):
        Example(**modify_kwargs(kwargs_valid, completions=_completions_all_empty))
    # completions - empty strings (1)
    if len(completions) > 1:
        _completions_with_empty = completions.copy()
        _completions_with_empty[0] = ""
        with pytest.raises(ValueError, match=re.escape("completions [0] are empty.")):
            Example(**modify_kwargs(kwargs_valid, completions=_completions_with_empty))

    # prior - not 1-D
    with pytest.raises(ValueError, match="prior must be 1-D."):
        Example(
            **modify_kwargs(
                kwargs_valid, prior=[[1 / len(completions)]] * len(completions)
            )
        )
    # prior - probabilities < 0
    with pytest.raises(
        ValueError, match="prior must contain probabilities between 0 and 1."
    ):
        Example(**modify_kwargs(kwargs_valid, prior=[-1] * len(completions)))
    # prior - probabilities > 1
    with pytest.raises(
        ValueError, match="prior must contain probabilities between 0 and 1."
    ):
        Example(**modify_kwargs(kwargs_valid, prior=[1.1] * len(completions)))



"""
Unit tests `cappr.utils._batch`.
"""

from __future__ import annotations
import re

import pytest

from cappr.utils import _batch


@pytest.fixture(scope="module")
def lst():
    return list(range(10))


def test_ProgressBar(lst: list):
    # test basic loop
    assert [x for x in _batch.ProgressBar(lst)] == lst
    assert [x for x in _batch.ProgressBar(lst, show_progress_bar=False)] == lst
    # test blank init like tqdm
    assert _batch.ProgressBar() is not None
    # test context manager
    lst_after_progress_bar = []
    with _batch.ProgressBar(total=len(lst), desc="context manager") as progress_bar:
        for x in lst:
            lst_after_progress_bar.append(x)
            progress_bar.update()


def _test_partition_and_order(batches: list[list], source_lst: list):
    """
    Asserts that `batches` partitions `source_lst` and are in the same order as in
    `source_lst`.
    """
    lst_from_batches = [x for batch in batches for x in batch]
    assert lst_from_batches == source_lst


@pytest.mark.parametrize("size", (-1, 2, 3))
def test_constant(lst: list, size: int):
    if size <= 0:
        with pytest.raises(ValueError, match="size must be positive"):
            next(_batch.constant(lst, size))
        return

    batches = list(_batch.constant(lst, size))

    _test_partition_and_order(batches, lst)

    # Test batch sizes
    for batch in batches[:-1]:
        assert len(batch) == size
    remaining = len(lst) % size
    expected_size_last = remaining if remaining else size
    assert len(batches[-1]) == expected_size_last


@pytest.mark.parametrize("sizes", ([2, 4, 3, 1], [10]))
def test_variable(lst, sizes):
    batches = list(_batch.variable(lst, sizes))

    _test_partition_and_order(batches, lst)

    assert len(batches) == len(sizes)

    # Test batch sizes
    for batch, expected_size in zip(batches, sizes):
        assert len(batch) == expected_size


def test_variable_bad_inputs():
    with pytest.raises(ValueError, match="sizes must all be positive"):
        next(_batch.variable([], sizes=[0, 1]))
    with pytest.raises(ValueError, match="sizes must be 1-D"):
        next(_batch.variable([], sizes=[[1]]))
    with pytest.raises(ValueError, match=re.escape("sizes must sum to len(lst)")):
        next(_batch.variable([1, 2, 3, 4], sizes=[1, 2]))


@pytest.mark.parametrize("batch_size", (2, 3))
def test_batchify(lst, batch_size):
    @_batch.batchify(batchable_arg="lst")
    def func(dummy_arg, lst, batch_size=100, dummy_kwarg="dummy", **kwargs):
        return lst

    out = func("dummy", lst, batch_size=batch_size)
    assert out == list(_batch.constant(lst, size=batch_size))
